# ***************************************************************************
# *   Copyright (c) 2020 Przemo Firszt <przemo@firszt.eu>                   *
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Analysis Checks"
__author__ = "Przemo Firszt, Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

import FreeCAD

from FreeCAD import Units

from . import femutils
from femsolver.calculix.solver import ANALYSIS_TYPES


def check_member_for_solver_calculix(analysis, solver, mesh, member):

    message = ""

    # solver
    if solver.AnalysisType not in ANALYSIS_TYPES:
        message += f"Unknown analysis type: {solver.AnalysisType}\n"
    if solver.AnalysisType == "frequency":
        if not hasattr(solver, "EigenmodeHighLimit"):
            message += "Frequency analysis: Solver has no EigenmodeHighLimit.\n"
        elif not hasattr(solver, "EigenmodeLowLimit"):
            message += "Frequency analysis: Solver has no EigenmodeLowLimit.\n"
        elif not hasattr(solver, "EigenmodesCount"):
            message += "Frequency analysis: Solver has no EigenmodesCount.\n"
    if hasattr(solver, "MaterialNonlinearity") and solver.MaterialNonlinearity == "nonlinear":
        if not member.mats_nonlinear:
            message += (
                "Solver is set to nonlinear materials, "
                "but there is no nonlinear material in the analysis.\n"
            )

    # mesh
    if not mesh:
        message += "A single mesh object must be defined in the analysis.\n"
    if mesh:
        if (
            mesh.FemMesh.VolumeCount == 0
            and mesh.FemMesh.FaceCount > 0
            and not member.geos_shellthickness
        ):
            message += (
                "FEM mesh has no volume elements, "
                "either define shell thicknesses or "
                "provide a FEM mesh with volume elements.\n"
            )
        if (
            mesh.FemMesh.VolumeCount == 0
            and mesh.FemMesh.FaceCount == 0
            and mesh.FemMesh.EdgeCount > 0
            and not member.geos_beamsection
            and not member.geos_fluidsection
        ):
            message += (
                "FEM mesh has no volume and no shell elements, "
                "either define a beam/fluid section or provide "
                "a FEM mesh with volume elements.\n"
            )
        if (
            mesh.FemMesh.VolumeCount == 0
            and mesh.FemMesh.FaceCount == 0
            and mesh.FemMesh.EdgeCount == 0
        ):
            message += (
                "FEM mesh has neither volume nor shell or edge elements. "
                "Provide a FEM mesh with elements.\n"
            )

    # material linear and nonlinear
    if not member.mats_linear:
        message += "No material object defined in the analysis.\n"
    has_no_references = False
    for m in member.mats_linear:
        if len(m["Object"].References) == 0:
            if has_no_references is True:
                message += (
                    "More than one material has an empty references list "
                    "(Only one empty references list is allowed!).\n"
                )
            has_no_references = True
    mat_ref_shty = ""
    for m in member.mats_linear:
        ref_shty = femutils.get_refshape_type(m["Object"])
        if ref_shty == "Compound":
            ref_shty = "Solid"
        if not mat_ref_shty:
            mat_ref_shty = ref_shty
        if mat_ref_shty and ref_shty and ref_shty != mat_ref_shty:
            # mat_ref_shty could be empty in one material
            # only the not empty ones should have the same shape type
            message += (
                "Some material objects do not have the same reference shape type "
                "(all material objects must have the same reference shape type, "
                "at the moment).\n"
            )
    for m in member.mats_linear:
        mat_map = m["Object"].Material
        mat_obj = m["Object"]
        if mat_obj.Category == "Solid":
            if "YoungsModulus" in mat_map:
                # print(Units.Quantity(mat_map["YoungsModulus"]).Value)
                if not Units.Quantity(mat_map["YoungsModulus"]).Value:
                    message += "Value of YoungsModulus is set to 0.0.\n"
            else:
                message += "No YoungsModulus defined for at least one material.\n"
            if "PoissonRatio" not in mat_map:
                # PoissonRatio is allowed to be 0.0 (in ccx), but it should be set anyway.
                message += "No PoissonRatio defined for at least one material.\n"
        if solver.AnalysisType == "frequency" or member.cons_selfweight:
            if "Density" not in mat_map:
                message += "No Density defined for at least one material.\n"
        if solver.AnalysisType == "thermomech":
            if "ThermalConductivity" in mat_map:
                if not Units.Quantity(mat_map["ThermalConductivity"]).Value:
                    message += "Value of ThermalConductivity is set to 0.0.\n"
            else:
                message += (
                    "Thermomechanical analysis: No ThermalConductivity defined "
                    "for at least one material.\n"
                )
            if "ThermalExpansionCoefficient" not in mat_map and mat_obj.Category == "Solid":
                message += (
                    "Thermomechanical analysis: No ThermalExpansionCoefficient defined "
                    "for at least one material.\n"  # allowed to be 0.0 (in ccx)
                )
            if "SpecificHeat" not in mat_map:
                message += (
                    "Thermomechanical analysis: No SpecificHeat "
                    "defined for at least one material.\n"  # allowed to be 0.0 (in ccx)
                )
        if femutils.is_of_type(mat_obj, "Fem::MaterialReinforced"):
            # additional tests for reinforced materials,
            # they are needed for result calculation, not for ccx analysis
            mat_map_m = mat_obj.Material
            if "AngleOfFriction" in mat_map_m:
                # print(Units.Quantity(mat_map_m["AngleOfFriction"]).Value)
                if not Units.Quantity(mat_map_m["AngleOfFriction"]).Value:
                    message += (
                        "Value of AngleOfFriction is set to 0.0 "
                        "for the matrix of a reinforced material.\n"
                    )
            else:
                message += (
                    "No AngleOfFriction defined for the matrix "
                    "of at least one reinforced material.\n"
                )
            if "CompressiveStrength" in mat_map_m:
                # print(Units.Quantity(mat_map_m["CompressiveStrength"]).Value)
                if not Units.Quantity(mat_map_m["CompressiveStrength"]).Value:
                    message += (
                        "Value of CompressiveStrength is set to 0.0 "
                        "for the matrix of a reinforced material.\n"
                    )
            else:
                message += (
                    "No CompressiveStrength defined for the matrix "
                    "of at least one reinforced material.\n"
                )
            mat_map_r = mat_obj.Reinforcement
            if "YieldStrength" in mat_map_r:
                # print(Units.Quantity(mat_map_r["YieldStrength"]).Value)
                if not Units.Quantity(mat_map_r["YieldStrength"]).Value:
                    message += (
                        "Value of YieldStrength is set to 0.0 "
                        "for the reinforcement of a reinforced material.\n"
                    )
            else:
                message += (
                    "No YieldStrength defined for the reinforcement "
                    "of at least one reinforced material.\n"
                )
    if len(member.mats_linear) == 1:
        mobj = member.mats_linear[0]["Object"]
        if hasattr(mobj, "References") and mobj.References:
            FreeCAD.Console.PrintError(
                "Only one material object, but this one has a reference shape. "
                "The reference shape will be ignored.\n"
            )
    for m in member.mats_linear:
        has_nonlinear_material = False
        for nlm in member.mats_nonlinear:
            if nlm["Object"].LinearBaseMaterial == m["Object"]:
                if has_nonlinear_material is False:
                    has_nonlinear_material = True
                else:
                    message += (
                        "At least two nonlinear materials use the same linear base material. "
                        "Only one nonlinear material for each linear material allowed.\n"
                    )

    # which analysis needs which constraints
    # no check in the regard of loads existence (constraint force, pressure, self weight)
    # is done, because an analysis without loads at all is a valid analysis too
    if solver.AnalysisType == "static":
        if not (member.cons_fixed or member.cons_displacement or member.cons_rigidbody):
            message += "Static analysis: No mechanical boundary conditions defined.\n"
    if solver.AnalysisType == "thermomech":
        if not member.cons_initialtemperature:
            if not member.geos_fluidsection:
                message += "Thermomechanical analysis: No initial temperature defined.\n"

    # constraints
    # fixed
    if member.cons_fixed:
        for c in member.cons_fixed:
            if len(c["Object"].References) == 0:
                message += "{} has empty references.".format(c["Object"].Name)
    # displacement
    if member.cons_displacement:
        for di in member.cons_displacement:
            if len(di["Object"].References) == 0:
                message += "{} has empty references.".format(c["Object"].Name)
    # plane rotation
    if member.cons_planerotation:
        for c in member.cons_planerotation:
            if len(c["Object"].References) == 0:
                message += "{} has empty references.".format(c["Object"].Name)
    # contact
    if member.cons_contact:
        for c in member.cons_contact:
            if len(c["Object"].References) == 0:
                message += "{} has empty references.".format(c["Object"].Name)
    # tie
    if member.cons_tie:
        for c in member.cons_tie:
            items = 0
            for reference in c["Object"].References:
                items += len(reference[1])
            if items != 2:
                message += "{} doesn't reference exactly two needed faces.\n".format(
                    c["Object"].Name
                )
    # sectionprint
    if member.cons_sectionprint:
        for c in member.cons_sectionprint:
            items = 0
            for reference in c["Object"].References:
                items += len(reference[1])
            if items != 1:
                message += "{} doesn't reference exactly one needed face.\n".format(
                    c["Object"].Name
                )
    # transform
    if member.cons_transform:
        for c in member.cons_transform:
            if len(c["Object"].References) == 0:
                message += "{} has empty references.".format(c["Object"].Name)
    # pressure
    if member.cons_pressure:
        for c in member.cons_pressure:
            if len(c["Object"].References) == 0:
                message += "{} has empty references.".format(c["Object"].Name)
    # force
    if member.cons_force:
        for c in member.cons_force:
            if len(c["Object"].References) == 0:
                message += "{} has empty references.".format(c["Object"].Name)
    # temperature
    if member.cons_temperature:
        for c in member.cons_temperature:
            if len(c["Object"].References) == 0:
                message += "{} has empty references.".format(c["Object"].Name)
    # heat flux
    if member.cons_heatflux:
        for c in member.cons_heatflux:
            if len(c["Object"].References) == 0:
                message += "{} has empty references.".format(c["Object"].Name)

    # geometries
    # beam section
    if member.geos_beamsection:
        if member.geos_shellthickness:
            # this needs to be checked only once either here or in shell_thicknesses
            message += (
                "Beam sections and shell thicknesses in one analysis "
                "are not supported at the moment.\n"
            )
        if member.geos_fluidsection:
            # this needs to be checked only once either here or in shell_thicknesses
            message += (
                "Beam sections and fluid sections in one analysis "
                "are not supported at the moment.\n"
            )
        has_no_references = False
        for b in member.geos_beamsection:
            if len(b["Object"].References) == 0:
                if has_no_references is True:
                    message += (
                        "More than one beam section has an empty references "
                        "list (Only one empty references list is allowed!).\n"
                    )
                has_no_references = True
        if mesh:
            if mesh.FemMesh.FaceCount > 0 or mesh.FemMesh.VolumeCount > 0:
                message += "Beam sections defined but FEM mesh has volume or shell elements.\n"
            if mesh.FemMesh.EdgeCount == 0:
                message += "Beam sections defined but FEM mesh has no edge elements.\n"
            if not (hasattr(mesh, "Shape") or hasattr(mesh, "Part")):
                message += (
                    "Mesh without geometry link. "
                    "The mesh needs to know its geometry for the beam rotations.\n"
                )
        if len(member.geos_beamrotation) > 1:
            message += "Multiple beam rotations in one analysis are not supported at the moment.\n"
    # beam rotations
    if member.geos_beamrotation and not member.geos_beamsection:
        message += "Beam rotations in the analysis but no beam sections defined.\n"
    # shell thickness
    if member.geos_shellthickness:
        has_no_references = False
        for s in member.geos_shellthickness:
            if len(s["Object"].References) == 0:
                if has_no_references is True:
                    message += (
                        "More than one shell thickness has an empty references "
                        "list (Only one empty references list is allowed!).\n"
                    )
                has_no_references = True
        if mesh:
            if mesh.FemMesh.VolumeCount > 0:
                message += "Shell thicknesses defined but FEM mesh has volume elements.\n"
            if mesh.FemMesh.FaceCount == 0:
                message += "Shell thicknesses defined but FEM mesh has no shell elements.\n"
    # fluid section
    if member.geos_fluidsection:
        if not member.cons_selfweight:
            message += "A fluid network analysis requires self weight constraint to be applied\n"
        if solver.AnalysisType != "thermomech":
            message += "A fluid network analysis can only be done in a thermomech analysis\n"
        has_no_references = False
        for f in member.geos_fluidsection:
            if len(f["Object"].References) == 0:
                if has_no_references is True:
                    message += (
                        "More than one fluid section has an empty references list "
                        "(Only one empty references list is allowed!).\n"
                    )
                has_no_references = True
        if mesh:
            if mesh.FemMesh.FaceCount > 0 or mesh.FemMesh.VolumeCount > 0:
                message += "Fluid sections defined but FEM mesh has volume or shell elements.\n"
            if mesh.FemMesh.EdgeCount == 0:
                message += "Fluid sections defined but FEM mesh has no edge elements.\n"

    return message


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import Path
import unittest


class TestPathLog(unittest.TestCase):
    """Some basic tests for the logging framework."""

    MODULE = "TestPathLog"  # file name without extension

    def setUp(self):
        Path.Log.setLevel(Path.Log.Level.RESET)
        Path.Log.untrackAllModules()

    def callerFile(self):
        return Path.Log._caller()[0]

    def callerLine(self):
        return Path.Log._caller()[1]

    def callerFunc(self):
        return Path.Log._caller()[2]

    def test00(self):
        """Check for proper module extraction."""
        self.assertEqual(self.callerFile(), self.MODULE)

    def test01(self):
        """Check for proper function extraction."""
        self.assertEqual(self.callerFunc(), "test01")

    def test10(self):
        """Verify default log levels is NOTICE."""
        self.assertEqual(Path.Log.getLevel(), Path.Log.Level.NOTICE)
        self.assertEqual(Path.Log.getLevel(self.MODULE), Path.Log.Level.NOTICE)

    def test11(self):
        """Verify setting global log level."""
        Path.Log.setLevel(Path.Log.Level.DEBUG)

        self.assertEqual(Path.Log.getLevel(), Path.Log.Level.DEBUG)
        self.assertEqual(Path.Log.getLevel(self.MODULE), Path.Log.Level.DEBUG)

    def test12(self):
        """Verify setting module log level."""
        Path.Log.setLevel(Path.Log.Level.DEBUG, self.MODULE)

        self.assertEqual(Path.Log.getLevel(), Path.Log.Level.NOTICE)
        self.assertEqual(Path.Log.getLevel(self.MODULE), Path.Log.Level.DEBUG)

    def test13(self):
        """Verify setting other modul's log level doesn't change this one's."""
        # if this test fails then most likely the global RESET is broken
        Path.Log.setLevel(Path.Log.Level.DEBUG, "SomeOtherModule")

        self.assertEqual(Path.Log.getLevel(), Path.Log.Level.NOTICE)
        self.assertEqual(Path.Log.getLevel(self.MODULE), Path.Log.Level.NOTICE)

    def test14(self):
        """Verify resetting log level for module falls back to global level."""
        Path.Log.setLevel(Path.Log.Level.DEBUG, self.MODULE)
        self.assertEqual(Path.Log.getLevel(self.MODULE), Path.Log.Level.DEBUG)
        # changing global log level does not affect module
        Path.Log.setLevel(Path.Log.Level.ERROR)
        self.assertEqual(Path.Log.getLevel(self.MODULE), Path.Log.Level.DEBUG)
        # resetting module log level restores global log level for module
        Path.Log.setLevel(Path.Log.Level.RESET, self.MODULE)
        self.assertEqual(Path.Log.getLevel(self.MODULE), Path.Log.Level.ERROR)
        # changing the global log level will also change the module log level
        Path.Log.setLevel(Path.Log.Level.DEBUG)
        self.assertEqual(Path.Log.getLevel(self.MODULE), Path.Log.Level.DEBUG)

    def test20(self):
        """Verify debug logs aren't logged by default."""
        self.assertIsNone(Path.Log.debug("this"))

    def test21(self):
        """Verify debug logs are logged if log level is set to DEBUG."""
        Path.Log.setLevel(Path.Log.Level.DEBUG)
        self.assertIsNotNone(Path.Log.debug("this"))

    def test30(self):
        """Verify log level ERROR."""
        Path.Log.setLevel(Path.Log.Level.ERROR)
        self.assertIsNone(Path.Log.debug("something"))
        self.assertIsNone(Path.Log.info("something"))
        self.assertIsNone(Path.Log.notice("something"))
        self.assertIsNone(Path.Log.warning("something"))
        self.assertIsNotNone(Path.Log.error("something"))

    def test31(self):
        """Verify log level WARNING."""
        Path.Log.setLevel(Path.Log.Level.WARNING)
        self.assertIsNone(Path.Log.debug("something"))
        self.assertIsNone(Path.Log.info("something"))
        self.assertIsNone(Path.Log.notice("something"))
        self.assertIsNotNone(Path.Log.warning("something"))
        self.assertIsNotNone(Path.Log.error("something"))

    def test32(self):
        """Verify log level NOTICE."""
        Path.Log.setLevel(Path.Log.Level.NOTICE)
        self.assertIsNone(Path.Log.debug("something"))
        self.assertIsNone(Path.Log.info("something"))
        self.assertIsNotNone(Path.Log.notice("something"))
        self.assertIsNotNone(Path.Log.warning("something"))
        self.assertIsNotNone(Path.Log.error("something"))

    def test33(self):
        """Verify log level INFO."""
        Path.Log.setLevel(Path.Log.Level.INFO)
        self.assertIsNone(Path.Log.debug("something"))
        self.assertIsNotNone(Path.Log.info("something"))
        self.assertIsNotNone(Path.Log.notice("something"))
        self.assertIsNotNone(Path.Log.warning("something"))
        self.assertIsNotNone(Path.Log.error("something"))

    def test34(self):
        """Verify log level DEBUG."""
        Path.Log.setLevel(Path.Log.Level.DEBUG)
        self.assertIsNotNone(Path.Log.debug("something"))
        self.assertIsNotNone(Path.Log.info("something"))
        self.assertIsNotNone(Path.Log.notice("something"))
        self.assertIsNotNone(Path.Log.warning("something"))
        self.assertIsNotNone(Path.Log.error("something"))

    def test50(self):
        """Verify no tracking by default."""
        self.assertIsNone(Path.Log.track("this", "and", "that"))

    def test51(self):
        """Verify enabling tracking for module results in tracking."""
        Path.Log.trackModule()
        # Don't want to rely on the line number matching - still want some
        # indication that track does the right thing ....
        msg = Path.Log.track("this", "and", "that")
        self.assertTrue(msg.startswith(self.MODULE))
        self.assertTrue(msg.endswith("test51(this, and, that)"))

    def test52(self):
        """Verify untracking stops tracking."""
        Path.Log.trackModule()
        self.assertIsNotNone(Path.Log.track("this", "and", "that"))
        Path.Log.untrackModule()
        self.assertIsNone(Path.Log.track("this", "and", "that"))

    def test53(self):
        """Verify trackAllModules works correctly."""
        Path.Log.trackAllModules(True)
        self.assertIsNotNone(Path.Log.track("this", "and", "that"))
        Path.Log.trackAllModules(False)
        self.assertIsNone(Path.Log.track("this", "and", "that"))
        Path.Log.trackAllModules(True)
        Path.Log.trackModule()
        self.assertIsNotNone(Path.Log.track("this", "and", "that"))
        Path.Log.trackAllModules(False)
        self.assertIsNotNone(Path.Log.track("this", "and", "that"))

    def test60(self):
        """Verify track handles no argument."""
        Path.Log.trackModule()
        msg = Path.Log.track()
        self.assertTrue(msg.startswith(self.MODULE))
        self.assertTrue(msg.endswith("test60()"))

    def test61(self):
        """Verify track handles arbitrary argument types correctly."""
        Path.Log.trackModule()
        msg = Path.Log.track("this", None, 1, 18.25)
        self.assertTrue(msg.startswith(self.MODULE))
        self.assertTrue(msg.endswith("test61(this, None, 1, 18.25)"))

    def testzz(self):
        """Restoring environment after tests."""
        Path.Log.setLevel(Path.Log.Level.RESET)



# ***************************************************************************
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Objects FEM"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

import FreeCAD


# PythonFeatures from package femobjects
# standard object name == class name == type without 'Fem::'

# PythonFeatures from package femsolver
# standard object name == type without 'Fem::'
# the class name is Proxy

# TODO
# There are objects which use a base object. It should be tested if the base object
# is in the same document as the doc in which the obj should be created.
# Could only be happen if the make is called from Python.
# What happens ATM? Error or the obj is moved to the other doc?


# ********* analysis objects *********************************************************************
def makeAnalysis(doc, name="Analysis"):
    """makeAnalysis(document, [name]):
    makes a Fem Analysis object"""
    obj = doc.addObject("Fem::FemAnalysis", name)
    return obj


# ********* constant objects *********************************************************************
def makeConstantVacuumPermittivity(doc, name="ConstantVacuumPermittivity"):
    """makeConstantVacuumPermittivity(document, [name]):
    makes a Fem ConstantVacuumPermittivity object"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constant_vacuumpermittivity

    constant_vacuumpermittivity.ConstantVacuumPermittivity(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constant_vacuumpermittivity

        view_constant_vacuumpermittivity.VPConstantVacuumPermittivity(obj.ViewObject)
    return obj


# ********* constraint objects *******************************************************************
def makeConstraintBearing(doc, name="ConstraintBearing"):
    """makeConstraintBearing(document, [name]):
    makes a Fem ConstraintBearing object"""
    obj = doc.addObject("Fem::ConstraintBearing", name)
    return obj


def makeConstraintBodyHeatSource(doc, name="ConstraintBodyHeatSource"):
    """makeConstraintBodyHeatSource(document, [name]):
    makes a Fem ConstraintBodyHeatSource object"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constraint_bodyheatsource

    constraint_bodyheatsource.ConstraintBodyHeatSource(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constraint_bodyheatsource as viewprov

        viewprov.VPConstraintBodyHeatSource(obj.ViewObject)
    return obj


def makeConstraintCentrif(doc, name="ConstraintCentrif"):
    """makeConstraintCentrif(document, [name]):
    creates a centrif object to define centrifugal body load constraint"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constraint_centrif

    constraint_centrif.ConstraintCentrif(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constraint_centrif

        view_constraint_centrif.VPConstraintCentrif(obj.ViewObject)
    return obj


def makeConstraintCurrentDensity(doc, name="ConstraintCurrentDensity"):
    """makeConstraintCurrentDensity(document, [name]):
    makes a Fem CurrentDensity object"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constraint_currentdensity

    constraint_currentdensity.ConstraintCurrentDensity(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constraint_currentdensity

        view_constraint_currentdensity.VPConstraintCurrentDensity(obj.ViewObject)
    return obj


def makeConstraintContact(doc, name="ConstraintContact"):
    """makeConstraintContact(document, [name]):
    makes a Fem ConstraintContact object"""
    obj = doc.addObject("Fem::ConstraintContact", name)
    return obj


def makeConstraintDisplacement(doc, name="ConstraintDisplacement"):
    """makeConstraintDisplacement(document, [name]):
    makes a Fem ConstraintDisplacement object"""
    obj = doc.addObject("Fem::ConstraintDisplacement", name)
    return obj


def makeConstraintElectricChargeDensity(doc, name="ElectricChargeDensity"):
    """makeConstraintElectricChargeDensity(document, [name]):
    makes a Fem ElectricChargeDensity object"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constraint_electricchargedensity

    constraint_electricchargedensity.ConstraintElectricChargeDensity(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constraint_electricchargedensity

        view_constraint_electricchargedensity.VPConstraintElectricChargeDensity(obj.ViewObject)
    return obj


def makeConstraintElectrostaticPotential(doc, name="ConstraintElectrostaticPotential"):
    """makeConstraintElectrostaticPotential(document, [name]):
    makes a Fem ElectrostaticPotential object"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constraint_electrostaticpotential

    constraint_electrostaticpotential.ConstraintElectrostaticPotential(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constraint_electrostaticpotential

        view_constraint_electrostaticpotential.VPConstraintElectroStaticPotential(obj.ViewObject)
    return obj


def makeConstraintFixed(doc, name="ConstraintFixed"):
    """makeConstraintFixed(document, [name]):
    makes a Fem ConstraintFixed object"""
    obj = doc.addObject("Fem::ConstraintFixed", name)
    return obj


def makeConstraintRigidBody(doc, name="ConstraintRigidBody"):
    """makeConstraintRigidBody(document, [name]):
    makes a Fem ConstraintRigidBody object"""
    obj = doc.addObject("Fem::ConstraintRigidBody", name)
    return obj


def makeConstraintFlowVelocity(doc, name="ConstraintFlowVelocity"):
    """makeConstraintFlowVelocity(document, [name]):
    makes a Fem ConstraintFlowVelocity object"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constraint_flowvelocity

    constraint_flowvelocity.ConstraintFlowVelocity(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constraint_flowvelocity

        view_constraint_flowvelocity.VPConstraintFlowVelocity(obj.ViewObject)
    return obj


def makeConstraintFluidBoundary(doc, name="ConstraintFluidBoundary"):
    """makeConstraintFluidBoundary(document, name):
    makes a Fem ConstraintFluidBoundary object"""
    obj = doc.addObject("Fem::ConstraintFluidBoundary", name)
    return obj


def makeConstraintForce(doc, name="ConstraintForce"):
    """makeConstraintForce(document, [name]):
    makes a Fem ConstraintForce object"""
    obj = doc.addObject("Fem::ConstraintForce", name)
    return obj


def makeConstraintGear(doc, name="ConstraintGear"):
    """makeConstraintGear(document, [name]):
    makes a Fem ConstraintGear object"""
    obj = doc.addObject("Fem::ConstraintGear", name)
    return obj


def makeConstraintHeatflux(doc, name="ConstraintHeatflux"):
    """makeConstraintHeatflux(document, [name]):
    makes a Fem ConstraintHeatflux object"""
    obj = doc.addObject("Fem::ConstraintHeatflux", name)
    return obj


def makeConstraintInitialFlowVelocity(doc, name="ConstraintInitialFlowVelocity"):
    """makeConstraintInitialFlowVelocity(document, [name]):
    makes a Fem ConstraintInitialFlowVelocity object"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constraint_initialflowvelocity

    constraint_initialflowvelocity.ConstraintInitialFlowVelocity(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constraint_initialflowvelocity

        view_constraint_initialflowvelocity.VPConstraintInitialFlowVelocity(obj.ViewObject)
    return obj


def makeConstraintInitialPressure(doc, name="ConstraintInitialPressure"):
    """makeConstraintInitialPressure(document, [name]):
    makes a Fem ConstraintInitialPressure object"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constraint_initialpressure

    constraint_initialpressure.ConstraintInitialPressure(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constraint_initialpressure

        view_constraint_initialpressure.VPConstraintInitialPressure(obj.ViewObject)
    return obj


def makeConstraintInitialTemperature(doc, name="ConstraintInitialTemperature"):
    """makeConstraintInitialTemperature(document, name):
    makes a Fem ConstraintInitialTemperature object"""
    obj = doc.addObject("Fem::ConstraintInitialTemperature", name)
    return obj


def makeConstraintMagnetization(doc, name="ConstraintMagnetization"):
    """makeConstraintMagnetization(document, [name]):
    makes a Fem Magnetization object"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constraint_magnetization

    constraint_magnetization.ConstraintMagnetization(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constraint_magnetization

        view_constraint_magnetization.VPConstraintMagnetization(obj.ViewObject)
    return obj


def makeConstraintPlaneRotation(doc, name="ConstraintPlaneRotation"):
    """makeConstraintPlaneRotation(document, [name]):
    makes a Fem ConstraintPlaneRotation object"""
    obj = doc.addObject("Fem::ConstraintPlaneRotation", name)
    return obj


def makeConstraintPressure(doc, name="ConstraintPressure"):
    """makeConstraintPressure(document, [name]):
    makes a Fem ConstraintPressure object"""
    obj = doc.addObject("Fem::ConstraintPressure", name)
    return obj


def makeConstraintPulley(doc, name="ConstraintPulley"):
    """makeConstraintPulley(document, [name]):
    makes a Fem ConstraintPulley object"""
    obj = doc.addObject("Fem::ConstraintPulley", name)
    return obj


def makeConstraintSelfWeight(doc, name="ConstraintSelfWeight"):
    """makeConstraintSelfWeight(document, [name]):
    creates a self weight object to define a gravity load"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constraint_selfweight

    constraint_selfweight.ConstraintSelfWeight(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constraint_selfweight

        view_constraint_selfweight.VPConstraintSelfWeight(obj.ViewObject)
    return obj


def makeConstraintTemperature(doc, name="ConstraintTemperature"):
    """makeConstraintTemperature(document, [name]):
    makes a Fem ConstraintTemperature object"""
    obj = doc.addObject("Fem::ConstraintTemperature", name)
    return obj


def makeConstraintTie(doc, name="ConstraintTie"):
    """makeConstraintTie(document, [name]):
    creates a tie object to define bonded faces constraint"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constraint_tie

    constraint_tie.ConstraintTie(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constraint_tie

        view_constraint_tie.VPConstraintTie(obj.ViewObject)
    return obj


def makeConstraintTransform(doc, name="ConstraintTransform"):
    """makeConstraintTransform(document, [name]):
    makes a Fem ConstraintTransform object"""
    obj = doc.addObject("Fem::ConstraintTransform", name)
    return obj


def makeConstraintSectionPrint(doc, name="ConstraintSectionPrint"):
    """makeConstraintSectionPrint(document, [name]):
    creates a section print object to evaluate forces and moments of defined face"""
    obj = doc.addObject("Fem::ConstraintPython", name)
    from femobjects import constraint_sectionprint

    constraint_sectionprint.ConstraintSectionPrint(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_constraint_sectionprint

        view_constraint_sectionprint.VPConstraintSectionPrint(obj.ViewObject)
    return obj


def makeConstraintSpring(doc, name="ConstraintSpring"):
    """makeConstraintSpring(document, [name]):
    makes a Fem ConstraintSpring object"""
    obj = doc.addObject("Fem::ConstraintSpring", name)
    return obj


# ********* element definition objects ***********************************************************
def makeElementFluid1D(doc, name="ElementFluid1D"):
    """makeElementFluid1D(document, [name]):
    creates a 1D fluid element object to define 1D flow"""
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import element_fluid1D

    element_fluid1D.ElementFluid1D(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_element_fluid1D

        view_element_fluid1D.VPElementFluid1D(obj.ViewObject)
    return obj


def makeElementGeometry1D(
    doc, sectiontype="Rectangular", width=10.0, height=25.0, thickness=2.0, name="ElementGeometry1D"
):
    """makeElementGeometry1D(document, [width], [height], [name]):
    creates a 1D geometry element object to define a cross section"""
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import element_geometry1D

    element_geometry1D.ElementGeometry1D(obj)

    obj.SectionType = sectiontype
    obj.RectWidth = width
    obj.RectHeight = height
    obj.CircDiameter = height
    obj.PipeDiameter = height
    obj.PipeThickness = thickness
    obj.Axis1Length = width
    obj.Axis2Length = height
    obj.BoxHeight = height
    obj.BoxWidth = width
    obj.BoxT1 = thickness
    obj.BoxT2 = thickness
    obj.BoxT3 = thickness
    obj.BoxT4 = thickness

    if FreeCAD.GuiUp:
        from femviewprovider import view_element_geometry1D

        view_element_geometry1D.VPElementGeometry1D(obj.ViewObject)
    return obj


def makeElementGeometry2D(doc, thickness=1.0, name="ElementGeometry2D"):
    """makeElementGeometry2D(document, [thickness], [name]):
    creates a 2D geometry element object to define a plate thickness"""
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import element_geometry2D

    element_geometry2D.ElementGeometry2D(obj)
    obj.Thickness = thickness
    if FreeCAD.GuiUp:
        from femviewprovider import view_element_geometry2D

        view_element_geometry2D.VPElementGeometry2D(obj.ViewObject)
    return obj


def makeElementRotation1D(doc, name="ElementRotation1D"):
    """makeElementRotation1D(document, [name]):
    creates a 1D geometry rotation element object to rotate a 1D cross section"""
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import element_rotation1D

    element_rotation1D.ElementRotation1D(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_element_rotation1D

        view_element_rotation1D.VPElementRotation1D(obj.ViewObject)
    return obj


# ********* material objects *********************************************************************
def makeMaterialFluid(doc, name="MaterialFluid"):
    """makeMaterialFluid(document, [name]):
    makes a FEM Material for fluid"""
    obj = doc.addObject("App::MaterialObjectPython", name)
    from femobjects import material_common

    material_common.MaterialCommon(obj)
    obj.Category = "Fluid"
    if FreeCAD.GuiUp:
        from femviewprovider import view_material_common

        view_material_common.VPMaterialCommon(obj.ViewObject)
    return obj


def makeMaterialMechanicalNonlinear(doc, base_material, name="MaterialMechanicalNonlinear"):
    """makeMaterialMechanicalNonlinear(document, base_material, [name]):
    creates a nonlinear material object"""
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import material_mechanicalnonlinear

    material_mechanicalnonlinear.MaterialMechanicalNonlinear(obj)
    obj.LinearBaseMaterial = base_material
    if FreeCAD.GuiUp:
        from femviewprovider import view_material_mechanicalnonlinear

        view_material_mechanicalnonlinear.VPMaterialMechanicalNonlinear(obj.ViewObject)
    return obj


def makeMaterialReinforced(doc, name="MaterialReinforced"):
    """makeMaterialReinforced(document, [matrix_material], [reinforcement_material], [name]):
    creates a reinforced material object"""
    obj = doc.addObject("App::MaterialObjectPython", name)
    from femobjects import material_reinforced

    material_reinforced.MaterialReinforced(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_material_reinforced

        view_material_reinforced.VPMaterialReinforced(obj.ViewObject)
    return obj


def makeMaterialSolid(doc, name="MaterialSolid"):
    """makeMaterialSolid(document, [name]):
    makes a FEM Material for solid"""
    obj = doc.addObject("App::MaterialObjectPython", name)
    from femobjects import material_common

    material_common.MaterialCommon(obj)
    obj.Category = "Solid"
    if FreeCAD.GuiUp:
        from femviewprovider import view_material_common

        view_material_common.VPMaterialCommon(obj.ViewObject)
    return obj


# ********* mesh objects *************************************************************************
def makeMeshBoundaryLayer(doc, base_mesh, name="MeshBoundaryLayer"):
    """makeMeshBoundaryLayer(document, base_mesh, [name]):
    creates a FEM mesh BoundaryLayer object to define boundary layer properties"""
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import mesh_boundarylayer

    mesh_boundarylayer.MeshBoundaryLayer(obj)
    # obj.BaseMesh = base_mesh
    # App::PropertyLinkList does not support append
    # we will use a temporary list to append the mesh BoundaryLayer obj. to the list
    tmplist = base_mesh.MeshBoundaryLayerList
    tmplist.append(obj)
    base_mesh.MeshBoundaryLayerList = tmplist
    if FreeCAD.GuiUp:
        from femviewprovider import view_mesh_boundarylayer

        view_mesh_boundarylayer.VPMeshBoundaryLayer(obj.ViewObject)
    return obj


def makeMeshGmsh(doc, name="MeshGmsh"):
    """makeMeshGmsh(document, [name]):
    makes a Gmsh FEM mesh object"""
    obj = doc.addObject("Fem::FemMeshShapeBaseObjectPython", name)
    from femobjects import mesh_gmsh

    mesh_gmsh.MeshGmsh(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_mesh_gmsh

        view_mesh_gmsh.VPMeshGmsh(obj.ViewObject)
    return obj


def makeMeshGroup(doc, base_mesh, use_label=False, name="MeshGroup"):
    """makeMeshGroup(document, base_mesh, [use_label], [name]):
    creates a FEM mesh refinement object to define properties for a region of a FEM mesh
    """
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import mesh_group

    mesh_group.MeshGroup(obj)
    obj.UseLabel = use_label
    # obj.BaseMesh = base_mesh
    # App::PropertyLinkList does not support append
    # we will use a temporary list to append the mesh group obj. to the list
    tmplist = base_mesh.MeshGroupList
    tmplist.append(obj)
    base_mesh.MeshGroupList = tmplist
    if FreeCAD.GuiUp:
        from femviewprovider import view_mesh_group

        view_mesh_group.VPMeshGroup(obj.ViewObject)
    return obj


def makeMeshNetgen(doc, name="MeshNetgen"):
    """makeMeshNetgen(document, [name]):
    makes a Netgen FEM mesh object"""
    obj = doc.addObject("Fem::FemMeshShapeBaseObjectPython", name)
    from femobjects import mesh_netgen

    mesh_netgen.MeshNetgen(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_mesh_netgen

        view_mesh_netgen.VPMeshNetgen(obj.ViewObject)
    return obj


def makeMeshNetgenLegacy(doc, name="MeshNetgen"):
    """makeMeshNetgenLegacy(document, [name]):
    makes a old implementation Netgen FEM mesh object"""
    obj = doc.addObject("Fem::FemMeshShapeNetgenObject", name)
    return obj


def makeMeshRegion(doc, base_mesh, element_length=0.0, name="MeshRegion"):
    """makeMeshRegion(document, base_mesh, [element_length], [name]):
    creates a FEM mesh refinement object to define properties for a refinement of a FEM mesh
    """
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import mesh_region

    mesh_region.MeshRegion(obj)
    obj.CharacteristicLength = element_length
    # obj.BaseMesh = base_mesh
    # App::PropertyLinkList does not support append
    # we will use a temporary list to append the mesh region obj. to the list
    tmplist = base_mesh.MeshRegionList
    tmplist.append(obj)
    base_mesh.MeshRegionList = tmplist
    if FreeCAD.GuiUp:
        from femviewprovider import view_mesh_region

        view_mesh_region.VPMeshRegion(obj.ViewObject)
    return obj


def makeMeshResult(doc, name="MeshResult"):
    """makeMeshResult(document, name): makes a Fem MeshResult object"""
    obj = doc.addObject("Fem::FemMeshObjectPython", name)
    from femobjects import mesh_result

    mesh_result.MeshResult(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_mesh_result

        view_mesh_result.VPFemMeshResult(obj.ViewObject)
    return obj


# ********* post processing objects **************************************************************
def makeResultMechanical(doc, name="ResultMechanical"):
    """makeResultMechanical(document, [name]):
    creates a mechanical result object to hold FEM results"""
    obj = doc.addObject("Fem::FemResultObjectPython", name)
    from femobjects import result_mechanical

    result_mechanical.ResultMechanical(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_result_mechanical

        view_result_mechanical.VPResultMechanical(obj.ViewObject)
    return obj


def makePostVtkFilterClipRegion(doc, base_vtk_result, name="VtkFilterClipRegion"):
    """makePostVtkFilterClipRegion(document, base_vtk_result, [name]):
    creates a FEM post processing region clip filter object (vtk based)"""
    obj = doc.addObject("Fem::FemPostClipFilter", name)
    base_vtk_result.addObject(obj)
    return obj


def makePostVtkFilterClipScalar(doc, base_vtk_result, name="VtkFilterClipScalar"):
    """makePostVtkFilterClipScalar(document, base_vtk_result, [name]):
    creates a FEM post processing scalar clip filter object (vtk based)"""
    obj = doc.addObject("Fem::FemPostScalarClipFilter", name)
    base_vtk_result.addObject(obj)
    return obj


def makePostVtkFilterCutFunction(doc, base_vtk_result, name="VtkFilterCutFunction"):
    """makePostVtkFilterCutFunction(document, base_vtk_result, [name]):
    creates a FEM post processing cut function filter object (vtk based)"""
    obj = doc.addObject("Fem::FemPostClipFilter", name)
    base_vtk_result.addObject(obj)
    return obj


def makePostVtkFilterWarp(doc, base_vtk_result, name="VtkFilterWarp"):
    """makePostVtkFilterWarp(document, base_vtk_result, [name]):
    creates a FEM post processing warp filter object (vtk based)"""
    obj = doc.addObject("Fem::FemPostWarpVectorFilter", name)
    base_vtk_result.addObject(obj)
    return obj


def makePostVtkFilterContours(doc, base_vtk_result, name="VtkFilterContours"):
    """makePostVtkFilterContours(document, base_vtk_result, [name]):
    creates a FEM post processing contours filter object (vtk based)"""
    obj = doc.addObject("Fem::FemPostContoursFilter", name)
    base_vtk_result.addObject(obj)
    return obj


def makePostFilterGlyph(doc, base_vtk_result, name="Glyph"):
    """makePostVtkFilterGlyph(document, [name]):
    creates a FEM post processing filter that visualizes vector fields with glyphs
    """
    obj = doc.addObject("Fem::PostFilterPython", name)
    from femobjects import post_glyphfilter

    post_glyphfilter.PostGlyphFilter(obj)
    base_vtk_result.addObject(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_post_glyphfilter

        view_post_glyphfilter.VPPostGlyphFilter(obj.ViewObject)
    return obj


def makePostVtkResult(doc, result_data, name="VtkResult"):
    """makePostVtkResult(document, base_result, [name]):
    creates a FEM post processing result data (vtk based) to hold FEM results
    Note: Result data get expanded, it can either be single result [result] or everything
          needed for a multistep result: [results_list, value_list, unit, description]
    """

    Pipeline_Name = "Pipeline_" + name
    obj = doc.addObject("Fem::FemPostPipeline", Pipeline_Name)
    obj.load(*result_data)
    if FreeCAD.GuiUp:
        obj.ViewObject.SelectionStyle = "BoundBox"
        # to assure the user sees something, set the default to Surface
        obj.ViewObject.DisplayMode = "Surface"
    return obj


def makePostLineplot(doc, name="Lineplot"):
    """makePostLineplot(document, [name]):
    creates a FEM post processing line plot
    """
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import post_lineplot

    post_lineplot.PostLineplot(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_post_lineplot

        view_post_lineplot.VPPostLineplot(obj.ViewObject)
    return obj


def makePostLineplotFieldData(doc, name="FieldData2D"):
    """makePostLineplotFieldData(document, [name]):
    creates a FEM post processing data extractor for 2D Field data
    """
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import post_lineplot

    post_lineplot.PostLineplotFieldData(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_post_lineplot

        view_post_lineplot.VPPostLineplotFieldData(obj.ViewObject)
    return obj


def makePostLineplotIndexOverFrames(doc, name="IndexOverFrames2D"):
    """makePostLineplotIndexOverFrames(document, [name]):
    creates a FEM post processing data extractor for 2D index data
    """
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import post_lineplot

    post_lineplot.PostLineplotIndexOverFrames(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_post_lineplot

        view_post_lineplot.VPPostLineplotIndexOverFrames(obj.ViewObject)
    return obj


def makePostHistogram(doc, name="Histogram"):
    """makePostHistogram(document, [name]):
    creates a FEM post processing histogram plot
    """
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import post_histogram

    post_histogram.PostHistogram(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_post_histogram

        view_post_histogram.VPPostHistogram(obj.ViewObject)
    return obj


def makePostHistogramFieldData(doc, name="FieldData1D"):
    """makePostHistogramFieldData(document, [name]):
    creates a FEM post processing data extractor for 1D Field data
    """
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import post_histogram

    post_histogram.PostHistogramFieldData(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_post_histogram

        view_post_histogram.VPPostHistogramFieldData(obj.ViewObject)
    return obj


def makePostHistogramIndexOverFrames(doc, name="IndexOverFrames1D"):
    """makePostHistogramIndexOverFrames(document, [name]):
    creates a FEM post processing data extractor for 1D Field data
    """
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import post_histogram

    post_histogram.PostHistogramIndexOverFrames(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_post_histogram

        view_post_histogram.VPPostHistogramIndexOverFrames(obj.ViewObject)
    return obj


def makePostTable(doc, name="Table"):
    """makePostTable(document, [name]):
    creates a FEM post processing histogram plot
    """
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import post_table

    post_table.PostTable(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_post_table

        view_post_table.VPPostTable(obj.ViewObject)
    return obj


def makePostTableFieldData(doc, name="FieldData1D"):
    """makePostTableFieldData(document, [name]):
    creates a FEM post processing data extractor for 1D Field data
    """
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import post_table

    post_table.PostTableFieldData(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_post_table

        view_post_table.VPPostTableFieldData(obj.ViewObject)
    return obj


def makePostTableIndexOverFrames(doc, name="IndexOverFrames1D"):
    """makePostTableIndexOverFrames(document, [name]):
    creates a FEM post processing data extractor for 1D Field data
    """
    obj = doc.addObject("Fem::FeaturePython", name)
    from femobjects import post_table

    post_table.PostTableIndexOverFrames(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_post_table

        view_post_table.VPPostTableIndexOverFrames(obj.ViewObject)
    return obj


# ********* solver objects ***********************************************************************
def _equation_creator(creator, base_solver, doc, name):
    eq = creator.create(doc, name)
    if base_solver:
        eq.Priority = 255 - len(base_solver.Group)
        base_solver.addObject(eq)
    return eq


def makeEquationDeformation(doc, base_solver=None, name="Deformation"):
    """makeEquationDeformation(document, [base_solver], [name]):
    creates a FEM deformation (nonlinear elasticity) equation for a solver"""
    from femsolver.elmer.equations import deformation

    return _equation_creator(deformation, base_solver, doc, name)


def makeEquationElasticity(doc, base_solver=None, name="Elasticity"):
    """makeEquationElasticity(document, [base_solver], [name]):
    creates a FEM elasticity equation for a solver"""
    from femsolver.elmer.equations import elasticity

    return _equation_creator(elasticity, base_solver, doc, name)


def makeEquationElectricforce(doc, base_solver=None, name="Electricforce"):
    """makeEquationElectricforce(document, [base_solver], [name]):
    creates a FEM Electricforce equation for a solver"""
    from femsolver.elmer.equations import electricforce

    return _equation_creator(electricforce, base_solver, doc, name)


def makeEquationElectrostatic(doc, base_solver=None, name="Electrostatic"):
    """makeEquationElectrostatic(document, [base_solver], [name]):
    creates a FEM electrostatic equation for a solver"""
    from femsolver.elmer.equations import electrostatic

    return _equation_creator(electrostatic, base_solver, doc, name)


def makeEquationFlow(doc, base_solver=None, name="Flow"):
    """makeEquationFlow(document, [base_solver], [name]):
    creates a FEM flow equation for a solver"""
    from femsolver.elmer.equations import flow

    return _equation_creator(flow, base_solver, doc, name)


def makeEquationFlux(doc, base_solver=None, name="Flux"):
    """makeEquationFlux(document, [base_solver], [name]):
    creates a FEM flux equation for a solver"""
    from femsolver.elmer.equations import flux

    return _equation_creator(flux, base_solver, doc, name)


def makeEquationHeat(doc, base_solver=None, name="Heat"):
    """makeEquationHeat(document, [base_solver], [name]):
    creates a FEM heat equation for a solver"""
    from femsolver.elmer.equations import heat

    return _equation_creator(heat, base_solver, doc, name)


def makeEquationMagnetodynamic(doc, base_solver=None, name="Magnetodynamic"):
    """makeEquationMagnetodynamic(document, [base_solver], [name]):
    creates a FEM magnetodynamic equation for a solver"""
    from femsolver.elmer.equations import magnetodynamic

    return _equation_creator(magnetodynamic, base_solver, doc, name)


def makeEquationMagnetodynamic2D(doc, base_solver=None, name="Magnetodynamic2D"):
    """makeEquationMagnetodynamic2D(document, [base_solver], [name]):
    creates a FEM magnetodynamic2D equation for a solver"""
    from femsolver.elmer.equations import magnetodynamic2D

    return _equation_creator(magnetodynamic2D, base_solver, doc, name)


def makeEquationStaticCurrent(doc, base_solver=None, name="StaticCurrent"):
    """makeEquationStaticCurrent(document, [base_solver], [name]):
    creates a FEM static current equation for a solver"""
    from femsolver.elmer.equations import staticcurrent

    return _equation_creator(staticcurrent, base_solver, doc, name)


def makeSolverCalculiXCcxTools(doc, name="SolverCcxTools"):
    """makeSolverCalculiXCcxTools(document, [name]):
    makes a Calculix solver object for the ccx tools module"""
    obj = doc.addObject("Fem::FemSolverObjectPython", name)
    from femobjects import solver_ccxtools

    solver_ccxtools.SolverCcxTools(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_solver_ccxtools

        view_solver_ccxtools.VPSolverCcxTools(obj.ViewObject)
    return obj


def makeSolverCalculiX(doc, name="SolverCalculiX"):
    """makeSolverCalculiX(document, [name]):
    makes a Calculix solver object"""
    obj = doc.addObject("Fem::FemSolverObjectPython", name)
    from femobjects import solver_calculix

    solver_calculix.SolverCalculiX(obj)
    if FreeCAD.GuiUp:
        from femviewprovider import view_solver_calculix

        view_solver_calculix.VPSolverCalculiX(obj.ViewObject)
    return obj


def makeSolverElmer(doc, name="SolverElmer"):
    """makeSolverElmer(document, [name]):
    makes a Elmer solver object"""
    obj = doc.addObject("Fem::FemSolverObjectPython", name)
    from femobjects import solver_elmer

    solver_elmer.SolverElmer(obj)
    obj.SimulationType = "Steady State"
    if FreeCAD.GuiUp:
        from femviewprovider import view_solver_elmer

        view_solver_elmer.VPSolverElmer(obj.ViewObject)
    return obj


def makeSolverMystran(doc, name="SolverMystran"):
    """makeSolverMystran(document, [name]):
    makes a Mystran solver object"""
    import femsolver.mystran.solver

    obj = femsolver.mystran.solver.create(doc, name)
    return obj


def makeSolverZ88(doc, name="SolverZ88"):
    """makeSolverZ88(document, [name]):
    makes a Z88 solver object"""
    import femsolver.z88.solver

    obj = femsolver.z88.solver.create(doc, name)
    return obj


"""
# get the supportedTypes
App.newDocument()
module = "Fem"
FreeCADGui.doCommand("import " + module)
for s in sorted(App.ActiveDocument.supportedTypes()):
    if s.startswith(module):
        print(s)

"""

##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2019 Zheng, Lei (realthunder)<realthunder.dev@gmail.com>*
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the base viewprovider code for the Link objects."""
## @package view_draftlink
# \ingroup draftviewproviders
# \brief Provides the base viewprovider code for the Link objects.

## \addtogroup draftviewproviders
# @{
from draftviewproviders.view_base import ViewProviderDraft


class ViewProviderDraftLink(ViewProviderDraft):
    """A view provider for link type object."""

    def __init__(self, vobj):
        super(ViewProviderDraftLink, self).__init__(vobj)

    def getIcon(self):
        tp = self.Object.Proxy.Type
        if tp == "Array":
            if self.Object.ArrayType == "ortho":
                return ":/icons/Draft_LinkArray.svg"
            elif self.Object.ArrayType == "polar":
                return ":/icons/Draft_PolarLinkArray.svg"
            elif self.Object.ArrayType == "circular":
                return ":/icons/Draft_CircularLinkArray.svg"
        elif tp == "PathArray":
            return ":/icons/Draft_PathLinkArray.svg"
        elif tp == "PathTwistedArray":
            return ":/icons/Draft_PathTwistedLinkArray.svg"
        elif tp == "PointArray":
            return ":/icons/Draft_PointLinkArray.svg"

    def claimChildren(self):
        obj = self.Object
        if hasattr(obj, "ExpandArray"):
            expand = obj.ExpandArray
        else:
            expand = obj.ShowElement
        if not expand:
            return super(ViewProviderDraftLink, self).claimChildren()
        else:
            return obj.ElementList


# Alias for compatibility with old versions of v0.19
_ViewProviderDraftLink = ViewProviderDraftLink

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2024 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to rotate objects in the 3D space."""
## @package gui_rotate
# \ingroup draftguitools
# \brief Provides GUI tools to rotate objects in the 3D space.

## \addtogroup draftguitools
# @{
import math
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import DraftVecUtils
from draftgeoutils import geometry
from draftguitools import gui_base_original
from draftguitools import gui_tool_utils
from draftguitools import gui_trackers as trackers
from draftutils import utils
from draftutils import todo
from draftutils.messages import _msg, _err, _toolmsg
from draftutils.translate import translate
from FreeCAD import Units as U


class Rotate(gui_base_original.Modifier):
    """Gui Command for the Rotate tool."""

    def GetResources(self):
        """Set icon, menu and tooltip."""
        return {
            "Pixmap": "Draft_Rotate",
            "Accel": "R, O",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Rotate", "Rotate"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Rotate",
                'Rotates the selected objects.\nIf the "Copy" option is active, it will create rotated copies.',
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated(name="Rotate")
        if not self.ui:
            return
        self.ghosts = []
        self.arctrack = None
        self.get_object_selection()

    def get_object_selection(self):
        """Get the object selection."""
        if Gui.Selection.hasSelection():
            return self.proceed()
        self.ui.selectUi(on_close_call=self.finish)
        _msg(translate("draft", "Select an object to rotate"))
        self.call = self.view.addEventCallback("SoEvent", gui_tool_utils.selectObject)

    def proceed(self):
        """Continue with the command after a selection has been made."""
        if self.call:
            self.view.removeEventCallback("SoEvent", self.call)
        self.selection = Gui.Selection.getSelectionEx("", 0)
        Gui.doCommand('selection = FreeCADGui.Selection.getSelectionEx("", 0)')
        self.step = 0
        self.center = None
        self.point = None
        self.firstangle = None
        self.ui.rotateSetCenterUi()
        self.arctrack = trackers.arcTracker()
        self.call = self.view.addEventCallback("SoEvent", self.action)
        _toolmsg(translate("draft", "Pick rotation center"))

    def action(self, arg):
        """Handle the 3D scene events.

        This is installed as an EventCallback in the Inventor view.

        Parameters
        ----------
        arg: dict
            Dictionary with strings that indicates the type of event received
            from the 3D view.
        """
        if arg["Type"] == "SoKeyboardEvent" and arg["Key"] == "ESCAPE":
            self.finish()
        elif not self.ui.mouse:
            pass
        elif arg["Type"] == "SoLocation2Event":
            self.handle_mouse_move_event(arg)
        elif (
            arg["Type"] == "SoMouseButtonEvent"
            and arg["State"] == "DOWN"
            and arg["Button"] == "BUTTON1"
        ):
            self.handle_mouse_click_event(arg)

    def _get_angle(self):
        if self.center is None:
            return 0
        if self.point is None:
            return 0
        if DraftVecUtils.dist(self.point, self.center) < 1e-7:
            return 0
        angle = DraftVecUtils.angle(self.wp.u, self.point.sub(self.center), self.wp.axis)
        if self.firstangle is None:
            return angle
        if angle < self.firstangle:
            return (2 * math.pi - self.firstangle) + angle
        return angle - self.firstangle

    def handle_mouse_move_event(self, arg):
        """Handle the mouse when moving."""
        for ghost in self.ghosts:
            ghost.off()
        self.point, ctrlPoint, info = gui_tool_utils.getPoint(self, arg)
        if self.center is not None:
            # Project self.point on a plane that is parallel to the wp and that
            # passes through self.center.
            self.point = geometry.project_point_on_plane(
                self.point, self.center, self.wp.axis, direction=None, force_projection=True
            )
        if self.extendedCopy:
            if not gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_alt_key()):
                self.step = 3
                self.finish()
        if self.step == 0:
            pass
        elif self.step == 1:
            angle = self._get_angle()
            self.ui.setRadiusValue(math.degrees(angle), unit="Angle")
            self.ui.radiusValue.setFocus()
            self.ui.radiusValue.selectAll()
        elif self.step == 2:
            angle = self._get_angle()
            self.arctrack.setApertureAngle(angle)
            for ghost in self.ghosts:
                ghost.rotate(self.wp.axis, angle)
                ghost.on()
            self.ui.setRadiusValue(math.degrees(angle), unit="Angle")
            self.ui.radiusValue.setFocus()
            self.ui.radiusValue.selectAll()
        gui_tool_utils.redraw3DView()

    def handle_mouse_click_event(self, arg):
        """Handle the mouse when the first button is clicked."""
        if not self.point:
            return
        if self.step == 0:
            self.set_center()
        elif self.step == 1:
            self.set_start_point()
        else:
            self.set_rotation_angle(arg)

    def set_center(self):
        """Set the center of the rotation."""
        if not self.ghosts:
            self.set_ghosts()
        self.center = self.point
        self.node = [self.point]
        self.ui.radiusUi()
        self.ui.radiusValue.setText(U.Quantity(0, U.Angle).UserString)
        self.ui.makeFace.hide()
        self.ui.labelRadius.setText(translate("draft", "Base angle"))
        self.ui.radiusValue.setToolTip(
            translate("draft", "The base angle to start the rotation from")
        )
        self.arctrack.setCenter(self.center)
        for ghost in self.ghosts:
            ghost.center(self.center)
        self.step = 1
        _toolmsg(translate("draft", "Pick base angle"))
        if self.planetrack:
            self.planetrack.set(self.point)

    def set_start_point(self):
        """Set the starting point of the rotation."""
        self.firstangle = self._get_angle()
        self.ui.labelRadius.setText(translate("draft", "Rotation"))
        self.ui.radiusValue.setToolTip(
            translate(
                "draft",
                "The amount of rotation to perform.\nThe final angle will be the base angle plus this amount.",
            )
        )
        self.arctrack.on()
        self.arctrack.setStartPoint(self.point)
        for ghost in self.ghosts:
            ghost.on()
        self.step = 2
        _toolmsg(translate("draft", "Pick rotation angle"))

    def set_rotation_angle(self, arg):
        """Set the rotation angle."""
        self.angle = self._get_angle()
        if self.angle != 0:
            self.rotate(
                self.ui.isCopy.isChecked()
                or gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_alt_key())
            )
        if gui_tool_utils.hasMod(arg, gui_tool_utils.get_mod_alt_key()):
            self.extendedCopy = True
        else:
            self.finish(cont=None)

    def set_ghosts(self):
        """Set the ghost to display."""
        for ghost in self.ghosts:
            ghost.remove()
        copy = self.ui.isCopy.isChecked()
        if self.ui.isSubelementMode.isChecked():
            self.ghosts = self.get_subelement_ghosts(self.selection, copy)
            if not self.ghosts:
                _err(translate("draft", "No valid subelements selected"))
        else:
            objs, places, _ = utils._modifiers_process_selection(
                self.selection, copy, add_movable_children=(not copy)
            )
            self.ghosts = [trackers.ghostTracker(objs, parent_places=places)]
        if self.center:
            for ghost in self.ghosts:
                ghost.center(self.center)

    def get_subelement_ghosts(self, selection, copy):
        """Get ghost for the subelements (vertices, edges)."""
        import Part

        ghosts = []
        for sel in selection:
            for sub in sel.SubElementNames if sel.SubElementNames else [""]:
                if (not copy and "Vertex" in sub) or "Edge" in sub:
                    shape = Part.getShape(sel.Object, sub, needSubElement=True, retType=0)
                    ghosts.append(trackers.ghostTracker(shape))
        return ghosts

    def finish(self, cont=False):
        """Terminate the operation.

        Parameters
        ----------
        cont: bool or None, optional
            Restart (continue) the command if `True`, or if `None` and
            `ui.continueMode` is `True`.
        """
        self.end_callbacks(self.call)
        if self.arctrack:
            self.arctrack.finalize()
        for ghost in self.ghosts:
            ghost.finalize()
        super().finish()
        if cont or (cont is None and self.ui and self.ui.continueMode):
            todo.ToDo.delayAfter(self.Activated, [])

    def rotate(self, copy):
        """Perform the rotation of the subelement(s) or the entire object(s)."""
        if copy:
            cmd_name = translate("draft", "Copy")
        else:
            cmd_name = translate("draft", "Rotate")
        Gui.addModule("Draft")
        cmd = "Draft.rotate(selection, "
        cmd += str(math.degrees(self.angle)) + ", "
        cmd += "center=" + DraftVecUtils.toString(self.center) + ", "
        cmd += "axis=" + DraftVecUtils.toString(self.wp.axis) + ", "
        cmd += "copy=" + str(copy) + ", "
        cmd += "subelements=" + str(self.ui.isSubelementMode.isChecked()) + ")"
        cmd_list = [cmd, "FreeCAD.ActiveDocument.recompute()"]
        self.commit(cmd_name, cmd_list)

    def numericInput(self, numx, numy, numz):
        """Validate the entry fields in the user interface.

        This function is called by the toolbar or taskpanel interface
        when valid x, y, and z have been entered in the input fields.
        """
        self.center = App.Vector(numx, numy, numz)
        self.node = [self.center]
        self.arctrack.setCenter(self.center)
        for ghost in self.ghosts:
            ghost.center(self.center)
        self.ui.radiusUi()
        self.ui.makeFace.hide()
        self.ui.labelRadius.setText(translate("draft", "Base angle"))
        self.ui.radiusValue.setToolTip(
            translate("draft", "The base angle to start the rotation from")
        )
        self.ui.radiusValue.setText(U.Quantity(0, U.Angle).UserString)
        self.step = 1
        _toolmsg(translate("draft", "Pick base angle"))

    def numericRadius(self, rad):
        """Validate the radius entry field in the user interface.

        This function is called by the toolbar or taskpanel interface
        when a valid radius has been entered in the input field.
        """
        if self.step == 1:
            self.ui.labelRadius.setText(translate("draft", "Rotation"))
            self.ui.radiusValue.setToolTip(
                translate(
                    "draft",
                    "The amount of rotation to perform.\nThe final angle will be the base angle plus this amount.",
                )
            )
            self.ui.radiusValue.setText(U.Quantity(0, U.Angle).UserString)
            self.firstangle = math.radians(rad)
            self.arctrack.setStartAngle(self.firstangle)
            self.arctrack.on()
            for ghost in self.ghosts:
                ghost.on()
            self.step = 2
            _toolmsg(translate("draft", "Pick rotation angle"))
        else:
            self.angle = math.radians(rad)
            self.rotate(self.ui.isCopy.isChecked())
            self.finish(cont=None)


Gui.addCommand("Draft_Rotate", Rotate())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Op.Gui.Base as PathOpGui
import Path.Op.Gui.PocketBase as PathPocketBaseGui
import Path.Op.Pocket as PathPocket

from PySide.QtCore import QT_TRANSLATE_NOOP

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


__title__ = "CAM Pocket Operation UI"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Pocket operation page controller and command implementation."


class TaskPanelOpPage(PathPocketBaseGui.TaskPanelOpPage):
    """Page controller class for Pocket operation"""

    def pocketFeatures(self):
        """pocketFeatures() ... return FeaturePocket (see PathPocketBaseGui)"""
        return PathPocketBaseGui.FeaturePocket | PathPocketBaseGui.FeatureRestMachining


Command = PathOpGui.SetupOperation(
    "Pocket3D",
    PathPocket.Create,
    TaskPanelOpPage,
    "CAM_3DPocket",
    QT_TRANSLATE_NOOP("CAM_Pocket3D", "3D Pocket"),
    QT_TRANSLATE_NOOP("CAM_Pocket3D", "Creates a 3D Pocket toolpath from a face or faces"),
    PathPocket.SetupProperties,
)

FreeCAD.Console.PrintLog("Loading PathPocketGui... done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2019, 2020 Carlo Pavan <carlopav@gmail.com>             *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to start the edit mode of different objects."""
## @package gui_edit
# \ingroup draftguitools
# \brief Provides GUI tools to start the edit mode of different objects.

__title__ = "FreeCAD Draft Edit Tool"
__author__ = (
    "Yorik van Havre, Werner Mayer, Martin Burbaum, Ken Cline, " "Dmitry Chigrin, Carlo Pavan"
)
__url__ = "https://www.freecad.org"

## \addtogroup draftguitools
# @{
import math
import pivy.coin as coin
import PySide.QtCore as QtCore
import PySide.QtWidgets as QtWidgets

import FreeCAD as App
import FreeCADGui as Gui
import DraftVecUtils
from draftutils import gui_utils
from draftutils import params
from draftutils import utils
from draftutils.translate import translate
from draftguitools import gui_base_original
from draftguitools import gui_edit_arch_objects as edit_arch
from draftguitools import gui_edit_draft_objects as edit_draft
from draftguitools import gui_edit_part_objects as edit_part
from draftguitools import gui_edit_sketcher_objects as edit_sketcher
from draftguitools import gui_tool_utils
from draftguitools import gui_trackers as trackers


COLORS = {
    "default": utils.get_rgba_tuple(params.get_param("snapcolor"))[:3],
    "black": (0.0, 0.0, 0.0),
    "white": (1.0, 1.0, 1.0),
    "grey": (0.5, 0.5, 0.5),
    "red": (1.0, 0.0, 0.0),
    "green": (0.0, 1.0, 0.0),
    "blue": (0.0, 0.0, 1.0),
    "yellow": (1.0, 1.0, 0.0),
    "cyan": (0.0, 1.0, 1.0),
    "magenta": (1.0, 0.0, 1.0),
}


class Edit(gui_base_original.Modifier):
    """The Draft_Edit FreeCAD command definition.

    A tool to graphically edit FreeCAD objects.
    Current implementation use many parts of pivy graphics code by user "looo".
    The tool collect editpoints from objects and display Trackers on them
    to allow editing their Shape and their parameters.

    Callbacks
    ---------
    selection_callback
        registered when tool is launched, identify
        selected objects.

    editing_callbacks
        self._keyPressedCB      -> self.keyPressed
        self._mouseMovedCB      -> self._mouseMovedCB
        if self._mousePressedCB -> self.mousePressed
        when trackers are displayed for selected objects,
        these callbacks capture user events and forward
        them to related functions

    Task panel (Draft Toolbar)
    ----------
    self.ui = Gui.draftToolBar

    Draft_Edit uses the taskpanel in 2 ways:

    1 - the user can select select an object and close the operation
        self.ui.editUi()

    2 - when editing, lineUi support clicking destination point
        by self.startEditing
        self.ui.lineUi()

    Tracker selection
    -----------------
    If the tool recognize mouse click as an attempt to startEditing,
    using soRayPickAction, it identifies the selected editTracker and
    start editing it. Here is where "looo" code was very useful.

    Editing preview
    ---------------
    When object editing begins, self.ghost is initiated with the
    corresponding DraftTracker of the object type. The object Tracker
    is deleted when user clicks again and endEditing.

    Context Menu
    ------------
    Activated with Alt+LeftClick or pressing key "e"
    It's a custom context menu, that depends on clicked tracker
    or on clicked object.

    display_tracker_menu
        populates the menu with custom actions

    evaluate_menu_action
        evaluate user chosen action and launch corresponding
        function.

    Preferences
    -----------
    max_objects: Int
        set by "DraftEditMaxObjects" in user preferences
        The max number of FreeCAD objects the tool is
        allowed to edit at the same time.

    pick_radius: Int
        set by "DraftEditPickRadius" in user preferences
        The pick radius during editing operation.
        Increase if you experience problems in clicking
        on a editTracker because of screen resolution.

    Attributes
    ----------
    obj: Edited object
        I'm planning to discard this attribute.
        In old implementation every function was supposed to
        act on self.obj, self.editpoints, self.trackers,
        self.pl, self.invpl.
        Due to multiple object editing, i'm planning to keep
        just self.trackers. Any other object will be identified
        and processed starting from editTracker information.

    editing: Int
        Index of the editTracker that has been clicked by the
        user. Tracker selection mechanism is based on it.
        if self.editing is None :
            the user didn't click any node, and next click will
            be processed as an attempt to start editing operation
        if self.editing == o or 1 or 2 or 3 etc :
            the user is editing corresponding node, so next click
            will be processed as an attempt to end editing operation

    trackers: Dictionary {object.Name : [editTrackers]}
        It records the list of DraftTrackers.editTracker.
        {object.Name as String : [editTrackers for the object]}
        Each tracker is created with (position,obj.Name,idx),
        so it's possible to recall it
        self.trackers[str(node.objectName.getValue())][ep]

    overNode: DraftTrackers.editTracker
        It represent the editTracker under the cursor position.
        It is used to preview the tracker selection action.

    ghost: DraftTrackers.*
        Handles the tracker to preview editing operations.
        it is initialized when user clicks on a editTracker
        by self.startEditing() function.

    alt_edit_mode: Int
        Allows alternative editing modes for objects.
        ATM supported for:
        - arcs: if 0 edit by 3 points, if 1 edit by center,
                radius, angles

    supportedObjs: List
        List of supported Draft Objects.
        The tool use utils.get_type(obj) to compare object type
        to the list.

    supportedCppObjs: List
        List of supported Part Objects.
        The tool use utils.get_type(obj) and obj.TypeId to compare
        object type to the list.
    """

    def __init__(self):
        super().__init__()
        """Initialize Draft_Edit Command."""
        self.running = False
        self.trackers = {"object": []}
        self.overNode = None  # preselected node with mouseover
        self.edited_objects = []
        self.obj = None
        self.editing = None

        # event callbacks
        self.selection_callback = None
        self._keyPressedCB = None
        self._mouseMovedCB = None
        self._mousePressedCB = None

        # this stores the DisplayMode of the object to restore it after editing
        # only used by Arch Structure
        self.objs_formats = {}

        # settings (get updated in Activated)
        self.max_objects = 5
        self.pick_radius = 20

        self.alt_edit_mode = 0  # default edit mode for objects

        # preview
        self.ghost = None

        # setup gui_tools for every supported object
        self.gui_tools_repository = GuiToolsRepository()

        self.gui_tools_repository.add("Wire", edit_draft.DraftWireGuiTools())
        self.gui_tools_repository.add("BSpline", edit_draft.DraftBSplineGuiTools())
        self.gui_tools_repository.add("BezCurve", edit_draft.DraftBezCurveGuiTools())
        self.gui_tools_repository.add("Circle", edit_draft.DraftCircleGuiTools())
        self.gui_tools_repository.add("Rectangle", edit_draft.DraftRectangleGuiTools())
        self.gui_tools_repository.add("Polygon", edit_draft.DraftPolygonGuiTools())
        self.gui_tools_repository.add("Ellipse", edit_draft.DraftEllipseGuiTools())
        self.gui_tools_repository.add(
            "Dimension", edit_draft.DraftDimensionGuiTools()
        )  # Backward compatibility
        self.gui_tools_repository.add("LinearDimension", edit_draft.DraftDimensionGuiTools())
        self.gui_tools_repository.add("Label", edit_draft.DraftLabelGuiTools())

        self.gui_tools_repository.add("Wall", edit_arch.ArchWallGuiTools())
        self.gui_tools_repository.add("Window", edit_arch.ArchWindowGuiTools())
        self.gui_tools_repository.add("Structure", edit_arch.ArchStructureGuiTools())
        self.gui_tools_repository.add("Space", edit_arch.ArchSpaceGuiTools())
        self.gui_tools_repository.add("PanelCut", edit_arch.ArchPanelCutGuiTools())
        self.gui_tools_repository.add("PanelSheet", edit_arch.ArchPanelSheetGuiTools())

        self.gui_tools_repository.add("Part::Line", edit_part.PartLineGuiTools())
        self.gui_tools_repository.add("Part::Box", edit_part.PartBoxGuiTools())
        self.gui_tools_repository.add("Part::Cylinder", edit_part.PartCylinderGuiTools())
        self.gui_tools_repository.add("Part::Cone", edit_part.PartConeGuiTools())
        self.gui_tools_repository.add("Part::Sphere", edit_part.PartSphereGuiTools())

        self.gui_tools_repository.add(
            "Sketcher::SketchObject", edit_sketcher.SketcherSketchObjectGuiTools()
        )

    def GetResources(self):
        return {
            "Pixmap": "Draft_Edit",
            "Accel": "D, E",
            "MenuText": QtCore.QT_TRANSLATE_NOOP("Draft_Edit", "Edit"),
            "ToolTip": QtCore.QT_TRANSLATE_NOOP("Draft_Edit", "Edits the active object"),
        }

    # -------------------------------------------------------------------------
    # MAIN FUNCTIONS
    # -------------------------------------------------------------------------

    def Activated(self):
        """
        Activated is run when user launch Edit command.
        If something is selected -> call self.proceed()
        If nothing is selected -> self.register_selection_callback()
        """
        if self.running:
            self.finish()
        super(Edit, self).Activated("Edit")
        if not App.ActiveDocument:
            self.finish()

        self.ui = Gui.draftToolBar
        self.view = gui_utils.get_3d_view()
        self.max_objects = params.get_param("DraftEditMaxObjects")
        self.pick_radius = params.get_param("DraftEditPickRadius")

        if Gui.Selection.getSelection():
            self.proceed()
        else:
            self.ui.selectUi(on_close_call=self.finish)
            App.Console.PrintMessage(translate("draft", "Select a Draft object to edit") + "\n")
            self.register_selection_callback()

    def proceed(self):
        """this method set the editTrackers"""
        self.unregister_selection_callback()
        self.edited_objects = self.getObjsFromSelection()
        if not self.edited_objects:
            return self.finish()

        self.format_objects_for_editing(self.edited_objects)

        # start object editing
        Gui.Selection.clearSelection()
        Gui.Snapper.setSelectMode(True)

        self.ui.editUi()

        for obj in self.edited_objects:
            self.setTrackers(obj, self.getEditPoints(obj))

        self.register_editing_callbacks()

    def numericInput(self, numx, numy, numz):
        """Execute callback by the toolbar to activate the update function.

        This function gets called by the toolbar
        or by the mouse click and activate the update function.
        """
        self.endEditing(self.obj, self.editing, App.Vector(numx, numy, numz))
        App.ActiveDocument.recompute()

    def finish(self, cont=False):
        """Terminate Edit Tool."""
        self.unregister_selection_callback()
        self.unregister_editing_callbacks()
        self.editing = None
        self.finalizeGhost()
        Gui.Snapper.setSelectMode(False)

        if self.ui:
            self.removeTrackers()

        if self.edited_objects:
            self.deformat_objects_after_editing(self.edited_objects)

        super(Edit, self).finish()
        self.running = False
        # delay resetting edit mode otherwise it doesn't happen
        from PySide import QtCore

        QtCore.QTimer.singleShot(0, self.reset_edit)

    def reset_edit(self):
        if Gui.ActiveDocument is not None:
            Gui.ActiveDocument.resetEdit()

    # -------------------------------------------------------------------------
    # SCENE EVENTS CALLBACKS
    # -------------------------------------------------------------------------

    def register_selection_callback(self):
        """Register callback for selection when command is launched."""
        self.unregister_selection_callback()
        self.selection_callback = self.view.addEventCallback("SoEvent", gui_tool_utils.selectObject)

    def unregister_selection_callback(self):
        """
        remove selection callback if it exists
        """
        try:
            if self.selection_callback:
                self.view.removeEventCallback("SoEvent", self.selection_callback)
        except RuntimeError:
            # the view has been deleted already
            pass
        self.selection_callback = None

    def register_editing_callbacks(self):
        """
        register editing callbacks (former action function)
        """
        self.render_manager = self.view.getViewer().getSoRenderManager()
        if self._keyPressedCB is None:
            self._keyPressedCB = self.view.addEventCallbackPivy(
                coin.SoKeyboardEvent.getClassTypeId(), self.keyPressed
            )
        if self._mouseMovedCB is None:
            self._mouseMovedCB = self.view.addEventCallbackPivy(
                coin.SoLocation2Event.getClassTypeId(), self.mouseMoved
            )
        if self._mousePressedCB is None:
            self._mousePressedCB = self.view.addEventCallbackPivy(
                coin.SoMouseButtonEvent.getClassTypeId(), self.mousePressed
            )
        # App.Console.PrintMessage("Draft edit callbacks registered \n")

    def unregister_editing_callbacks(self):
        """
        remove callbacks used during editing if they exist
        """
        try:
            if self._keyPressedCB:
                self.view.removeEventCallbackSWIG(
                    coin.SoKeyboardEvent.getClassTypeId(), self._keyPressedCB
                )
                # App.Console.PrintMessage("Draft edit keyboard callback unregistered \n")
            if self._mouseMovedCB:
                self.view.removeEventCallbackSWIG(
                    coin.SoLocation2Event.getClassTypeId(), self._mouseMovedCB
                )
                # App.Console.PrintMessage("Draft edit location callback unregistered \n")
            if self._mousePressedCB:
                self.view.removeEventCallbackSWIG(
                    coin.SoMouseButtonEvent.getClassTypeId(), self._mousePressedCB
                )
                # App.Console.PrintMessage("Draft edit mouse button callback unregistered \n")
        except RuntimeError:
            # the view has been deleted already
            pass
        self._keyPressedCB = None
        self._mouseMovedCB = None
        self._mousePressedCB = None

    # -------------------------------------------------------------------------
    # SCENE EVENT HANDLERS
    # -------------------------------------------------------------------------

    def keyPressed(self, event_callback):
        """Execute as callback for keyboard event."""
        # TODO: Get the keys from preferences
        event = event_callback.getEvent()
        if event.getState() in (coin.SoKeyboardEvent.DOWN, coin.SoKeyboardEvent.UP):
            key = event.getKey()
            # App.Console.PrintMessage("pressed key : "+str(key)+"\n")
            if key == 65307:  # ESC
                self.finish()
            if key == 101:  # "e"
                self.display_tracker_menu(event)
            if (
                key == 65535 and Gui.Selection.getSelection() is None
            ):  # BUG: delete key activate Std::Delete command at the same time!
                print("DELETE PRESSED\n")
                self.delPoint(event)

    def mousePressed(self, event_callback):
        """
        mouse button event handler, calls: startEditing, endEditing, addPoint, delPoint
        """
        event = event_callback.getEvent()
        if (
            event.getState() == coin.SoMouseButtonEvent.DOWN and event.getButton() == event.BUTTON1
        ):  # left click
            if not event.wasAltDown():
                if self.editing is None:

                    pos = event.getPosition()
                    node = self.getEditNode(pos)
                    node_idx = self.getEditNodeIndex(node)
                    if node_idx is None:
                        return
                    doc = App.getDocument(str(node.documentName.getValue()))
                    obj = doc.getObject(str(node.objectName.getValue()))

                    self.startEditing(obj, node_idx)
                else:
                    self.endEditing(self.obj, self.editing)
            elif event.wasAltDown():  # left click with ctrl down
                self.display_tracker_menu(event)

    def mouseMoved(self, event_callback):
        """Execute as callback for mouse movement.

        Update tracker position and update preview ghost.
        """
        event = event_callback.getEvent()
        if self.editing is not None:
            self.updateTrackerAndGhost(event)
        else:
            # look for a node in mouse position and highlight it
            pos = event.getPosition()
            node = self.getEditNode(pos)
            ep = self.getEditNodeIndex(node)
            if ep is not None:
                if self.overNode is not None:
                    self.overNode.setColor(COLORS["default"])
                self.trackers[str(node.objectName.getValue())][ep].setColor(COLORS["red"])
                self.overNode = self.trackers[str(node.objectName.getValue())][ep]
            else:
                if self.overNode is not None:
                    self.overNode.setColor(COLORS["default"])
                    self.overNode = None

    def startEditing(self, obj, node_idx):
        """Start editing selected EditNode."""
        self.obj = obj  # this is still needed to handle preview
        if obj is None:
            return

        App.Console.PrintMessage(obj.Name + ": editing node number " + str(node_idx) + "\n")

        self.ui.lineUi(title=translate("draft", "Edit Node"), icon="Draft_Edit")
        self.ui.continueCmd.hide()
        self.editing = node_idx
        self.trackers[obj.Name][node_idx].off()

        self.finalizeGhost()
        self.initGhost(obj)

        self.node.append(self.trackers[obj.Name][node_idx].get())
        Gui.Snapper.setSelectMode(False)
        self.hideTrackers()

    def updateTrackerAndGhost(self, event):
        """Update tracker position when editing and update ghost."""
        pos = event.getPosition().getValue()
        orthoConstrain = False
        if event.wasShiftDown() == 1:
            orthoConstrain = True
        snappedPos = Gui.Snapper.snap((pos[0], pos[1]), self.node[-1], constrain=orthoConstrain)
        self.trackers[self.obj.Name][self.editing].set(snappedPos)
        self.ui.displayPoint(snappedPos, self.node[-1], mask=Gui.Snapper.affinity)
        if self.ghost:
            self.updateGhost(obj=self.obj, node_idx=self.editing, v=snappedPos)

    def endEditing(self, obj, nodeIndex, v=None):
        """Terminate editing and start object updating process."""
        self.finalizeGhost()
        self.trackers[obj.Name][nodeIndex].on()
        Gui.Snapper.setSelectMode(True)
        if v is None:
            # endEditing is called by mousePressed
            v = self.trackers[obj.Name][nodeIndex].get()
        else:
            # endEditing is called by numericInput, so tracker
            # position should be updated manually
            self.trackers[obj.Name][nodeIndex].set(v)
        self.update(obj, nodeIndex, v)
        self.alt_edit_mode = 0
        self.ui.editUi()
        self.node = []
        self.editing = None
        self.showTrackers()
        gui_tool_utils.redraw_3d_view()

    # -------------------------------------------------------------------------
    # EDIT TRACKERS functions
    # -------------------------------------------------------------------------

    def setTrackers(self, obj, points=None):
        """Set Edit Trackers for editpoints collected from self.obj."""
        if utils.get_type(obj) == "BezCurve":
            return self.resetTrackersBezier(obj)
        if points is None or len(points) == 0:
            _wrn = translate("draft", "No edit point found for selected object")
            App.Console.PrintWarning(_wrn + "\n")
            # do not finish if some trackers are still present
            if self.trackers == {"object": []}:
                self.finish()
            return
        self.trackers[obj.Name] = []
        if obj.Name in self.trackers:
            self.removeTrackers(obj)
        for ep in range(len(points)):
            self.trackers[obj.Name].append(
                trackers.editTracker(pos=points[ep], name=obj.Name, idx=ep)
            )

    def resetTrackers(self, obj):
        """Reset Edit Trackers and set them again."""
        self.removeTrackers(obj)
        self.setTrackers(obj, self.getEditPoints(obj))

    def resetTrackersBezier(self, obj):
        # in future move tracker definition to DraftTrackers
        size = params.get_param_view("MarkerSize")
        knotmarkers = (
            Gui.getMarkerIndex("DIAMOND_FILLED", size),  # sharp
            Gui.getMarkerIndex("SQUARE_FILLED", size),  # tangent
            Gui.getMarkerIndex("HOURGLASS_FILLED", size),  # symmetric
        )
        polemarker = Gui.getMarkerIndex("CIRCLE_FILLED", size)  # pole
        self.trackers[obj.Name] = []
        cont = obj.Continuity
        firstknotcont = cont[-1] if (obj.Closed and cont) else 0
        pointswithmarkers = [(obj.Shape.Edges[0].Curve.getPole(1), knotmarkers[firstknotcont])]
        for edgeindex, edge in enumerate(obj.Shape.Edges):
            poles = edge.Curve.getPoles()
            pointswithmarkers.extend([(point, polemarker) for point in poles[1:-1]])
            if not obj.Closed or len(obj.Shape.Edges) > edgeindex + 1:
                knotmarkeri = cont[edgeindex] if len(cont) > edgeindex else 0
                pointswithmarkers.append((poles[-1], knotmarkers[knotmarkeri]))
        for index, pwm in enumerate(pointswithmarkers):
            p, marker = pwm
            p = obj.Placement.inverse().multVec(p)
            p = obj.getGlobalPlacement().multVec(p)
            self.trackers[obj.Name].append(
                trackers.editTracker(p, obj.Name, index, obj.ViewObject.LineColor, marker=marker)
            )

    def removeTrackers(self, obj=None):
        """Remove Edit Trackers.

        Attributes
        ----------
        obj: FreeCAD object
            Removes trackers only for given object,
            if obj is None, removes all trackers
        """
        if obj is None:
            for key in self.trackers:
                for t in self.trackers[key]:
                    t.finalize()
            self.trackers = {"object": []}
        else:
            key = obj.Name
            if key in self.trackers:
                for t in self.trackers[key]:
                    t.finalize()
            self.trackers[key] = []

    def hideTrackers(self, obj=None):
        """Hide Edit Trackers.

        Attributes
        ----------
        obj: FreeCAD object
            Hides trackers only for given object,
            if obj is None, hides all trackers
        """
        if obj is None:
            for key in self.trackers:
                for t in self.trackers[key]:
                    t.off()
        else:
            for t in self.trackers[obj.Name]:
                t.off()

    def showTrackers(self, obj=None):
        """Show Edit Trackers.

        Attributes
        ----------
        obj: FreeCAD object
            Shows trackers only for given object,
            if obj is None, shows all trackers
        """
        if obj is None:
            for key in self.trackers:
                for t in self.trackers[key]:
                    t.on()
        else:
            for t in self.trackers[obj.Name]:
                t.on()

    # -------------------------------------------------------------------------
    # PREVIEW
    # -------------------------------------------------------------------------

    def initGhost(self, obj):
        self.current_editing_object_gui_tools = self.get_obj_gui_tools(obj)
        if self.current_editing_object_gui_tools:
            self.ghost = self.current_editing_object_gui_tools.init_preview_object(obj)

    def updateGhost(self, obj, node_idx, v):
        self.ghost.on()

        if self.current_editing_object_gui_tools:
            self.current_editing_object_gui_tools.update_preview_object(self, obj, node_idx, v)

        gui_tool_utils.redraw_3d_view()

    def finalizeGhost(self):
        try:
            self.current_editing_object_gui_tools = None
            self.ghost.finalize()
            self.ghost = None
        except Exception:
            return

    # ------------------------------------------------------------------------
    # DRAFT EDIT Context menu
    # ------------------------------------------------------------------------

    def display_tracker_menu(self, event):
        self.tracker_menu = QtWidgets.QMenu()
        actions = None

        if self.overNode:
            # if user is over a node
            doc = self.overNode.get_doc_name()
            obj = App.getDocument(doc).getObject(self.overNode.get_obj_name())
            ep = self.overNode.get_subelement_index()

            obj_gui_tools = self.get_obj_gui_tools(obj)
            if obj_gui_tools:
                actions = obj_gui_tools.get_edit_point_context_menu(self, obj, ep)

        else:
            # try if user is over an edited object
            pos = event.getPosition().getValue()
            obj = self.get_selected_obj_at_position(pos)

            obj_gui_tools = self.get_obj_gui_tools(obj)
            if obj_gui_tools:
                actions = obj_gui_tools.get_edit_obj_context_menu(self, obj, pos)

        if actions is None:
            return

        for label, callback in actions:

            def wrapper(callback=callback):
                callback()
                self.resetTrackers(obj)

            action = self.tracker_menu.addAction(label)
            action.setData(wrapper)

        self.tracker_menu.popup(Gui.getMainWindow().cursor().pos())

        QtCore.QObject.connect(
            self.tracker_menu, QtCore.SIGNAL("triggered(QAction *)"), self.evaluate_menu_action
        )

    def evaluate_menu_action(self, action):
        callback = action.data()
        callback()

    # -------------------------------------------------------------------------
    # EDIT OBJECT TOOLS
    #
    # This section contains the code to retrieve the object points and update them
    #
    # -------------------------------------------------------------------------

    def getEditPoints(self, obj):
        """Return a list of App.Vectors according to the given object edit nodes."""
        eps = None

        obj_gui_tools = self.get_obj_gui_tools(obj)
        if obj_gui_tools:
            eps = obj_gui_tools.get_edit_points(obj)

        if eps:
            return self.globalize_vectors(obj, eps)
        else:
            return None

    def update(self, obj, nodeIndex, v):
        """Apply the App.Vector to the modified point and update obj."""
        v = self.localize_vector(obj, v)
        App.ActiveDocument.openTransaction("Edit")
        self.update_object(obj, nodeIndex, v)
        App.ActiveDocument.commitTransaction()
        self.resetTrackers(obj)
        try:
            gui_tool_utils.redraw_3d_view()
        except AttributeError as err:
            pass

    def update_object(self, obj, nodeIndex, v):
        """Update the object according to the given modified editpoint."""
        obj_gui_tools = self.get_obj_gui_tools(obj)
        if obj_gui_tools:
            eps = obj_gui_tools.update_object_from_edit_points(
                obj, nodeIndex, v, self.alt_edit_mode
            )

        obj.recompute()

    # -------------------------------------------------------------------------
    # UTILS
    # -------------------------------------------------------------------------

    def has_obj_gui_tools(self, obj):
        """Check if the object has the GuiTools to provide information to edit it."""
        if (
            hasattr(obj, "obj_gui_tools")
            or (hasattr(obj, "Proxy") and hasattr(obj.Proxy, "obj_gui_tools"))
            or (utils.get_type(obj) in self.gui_tools_repository.keys())
        ):
            return True
        else:
            return False

    def get_obj_gui_tools(self, obj):
        """Retrieve the obj_gui_tools to support Draft Edit."""
        try:
            obj_gui_tools = obj.obj_gui_tools
        except AttributeError:
            try:
                obj_gui_tools = obj.Proxy.obj_gui_tools
            except AttributeError:
                try:
                    obj_gui_tools = self.gui_tools_repository.get(utils.get_type(obj))
                except Exception:
                    obj_gui_tools = None
        return obj_gui_tools

    def getObjsFromSelection(self):
        """Evaluate selection and return a valid object to edit.

        #to be used for app link support

        for selobj in Gui.Selection.getSelectionEx('', 0):
            for sub in selobj.SubElementNames:
                obj = selobj.Object
                obj_matrix = selobj.Object.getSubObject(sub, retType=4)
        """
        selection = Gui.Selection.getSelection()
        self.edited_objects = []
        if len(selection) > self.max_objects:
            _err = translate("draft", "Too many objects selected, maximum number set to:")
            App.Console.PrintMessage(_err + " " + str(self.max_objects) + "\n")
            return None

        for obj in selection:
            if self.has_obj_gui_tools(obj):
                self.edited_objects.append(obj)
            else:
                _wrn = translate("draft", ": this object is not editable")
                App.Console.PrintWarning(obj.Name + _wrn + "\n")
        return self.edited_objects

    def format_objects_for_editing(self, objs):
        """Change objects style during editing mode."""
        for obj in objs:
            obj_gui_tools = self.get_obj_gui_tools(obj)
            if obj_gui_tools:
                self.objs_formats[obj.Name] = obj_gui_tools.get_object_style(obj)
                obj_gui_tools.set_object_editing_style(obj)

    def deformat_objects_after_editing(self, objs):
        """Restore objects style during editing mode."""
        for obj in objs:
            if utils.is_deleted(obj):
                continue
            obj_gui_tools = self.get_obj_gui_tools(obj)
            if obj_gui_tools:
                obj_gui_tools.restore_object_style(obj, self.objs_formats[obj.Name])

    def get_specific_object_info(self, obj, pos):
        """Return info of a specific object at a given position."""
        selobjs = self.view.getObjectsInfo((pos[0], pos[1]))
        if not selobjs:
            return
        for info in selobjs:
            if not info:
                continue
            if obj.Name == info["Object"] and "x" in info:
                # prefer "real" 3D location over working-plane-driven one if possible
                pt = App.Vector(info["x"], info["y"], info["z"])
                return info, pt

    def get_selected_obj_at_position(self, pos):
        """Return object at given position.

        If object is one of the edited objects (self.edited_objects).
        """
        selobjs = self.view.getObjectsInfo((pos[0], pos[1]))
        if not selobjs:
            return
        for info in selobjs:
            if not info:
                return
            for obj in self.edited_objects:
                if obj.Name == info["Object"]:
                    return obj

    def globalize_vectors(self, obj, pointList):
        """Return the given point list in the global coordinate system."""
        plist = []
        for p in pointList:
            point = self.globalize_vector(obj, p)
            plist.append(point)
        return plist

    def globalize_vector(self, obj, point):
        """Return the given point in the global coordinate system."""
        if hasattr(obj, "getGlobalPlacement"):
            return obj.getGlobalPlacement().multVec(point)
        else:
            return point

    def localize_vectors(self, obj, pointList):
        """Return the given point list in the given object coordinate system."""
        plist = []
        for p in pointList:
            point = self.localize_vector(obj, p)
            plist.append(point)
        return plist

    def localize_vector(self, obj, point):
        """Return the given point in the given object coordinate system."""
        if hasattr(obj, "getGlobalPlacement"):
            return obj.getGlobalPlacement().inverse().multVec(point)
        else:
            return point

    def getEditNode(self, pos):
        """Get edit node from given screen position."""
        node = self.sendRay(pos)
        return node

    def sendRay(self, mouse_pos):
        """Send a ray through the scene and return the nearest entity."""
        ray_pick = coin.SoRayPickAction(self.render_manager.getViewportRegion())
        ray_pick.setPoint(coin.SbVec2s(*mouse_pos))
        ray_pick.setRadius(self.pick_radius)
        ray_pick.setPickAll(True)
        ray_pick.apply(self.render_manager.getSceneGraph())
        picked_point = ray_pick.getPickedPointList()
        return self.searchEditNode(picked_point)

    def searchEditNode(self, picked_point):
        """Search edit node inside picked point list and return node number."""
        for point in picked_point:
            path = point.getPath()
            length = path.getLength()
            point = path.getNode(length - 2)
            # import DraftTrackers
            if hasattr(point, "subElementName") and "EditNode" in str(
                point.subElementName.getValue()
            ):
                return point
        return None

    def getEditNodeIndex(self, point):
        """Get edit node index from given screen position."""
        if point:
            subElement = str(point.subElementName.getValue())
            ep = int(subElement[8:])
            return ep
        else:
            return None


class GuiToolsRepository:
    """This object provide a repository to collect all the specific objects
    editing tools.
    """

    def __init__(self):
        self.obj_gui_tools = {}

    def get(self, obj_type):
        return self.obj_gui_tools[obj_type]

    def add(self, type, gui_tools):
        self.obj_gui_tools[type] = gui_tools

    def keys(self):
        return self.obj_gui_tools.keys()


Gui.addCommand("Draft_Edit", Edit())

## @}



# ***************************************************************************
# *   Copyright (c) 2014 Johan Kristensen                                   *
# *   Copyright (c) 2014 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD, FreeCADGui, Sketcher, Part, math


__title__ = "Regular polygon profile lib"
__author__ = "Johan Kristensen"
__url__ = "https://www.freecad.org"

App = FreeCAD
Gui = FreeCADGui


def makeRegularPolygon(
    sketch,
    sides,
    centerPoint=App.Vector(0, 0, 0),
    firstCornerPoint=App.Vector(-20.00, 34.64, 0),
    construction=False,
):

    if not sketch:
        App.Console.PrintError("No sketch specified in 'makeRegularPolygon'")
        return
    if sides < 3:
        App.Console.PrintError("Number of sides must be at least 3 in 'makeRegularPolygon'")
        return

    diffVec = firstCornerPoint - centerPoint
    diffVec.z = 0
    angular_diff = 2 * math.pi / sides
    pointList = []
    for i in range(0, sides):
        cos_v = math.cos(angular_diff * i)
        sin_v = math.sin(angular_diff * i)
        pointList.append(
            centerPoint
            + App.Vector(
                cos_v * diffVec.x - sin_v * diffVec.y, cos_v * diffVec.y + sin_v * diffVec.x, 0
            )
        )

    geoList = []
    for i in range(0, sides - 1):
        geoList.append(Part.LineSegment(pointList[i], pointList[i + 1]))
    geoList.append(Part.LineSegment(pointList[sides - 1], pointList[0]))
    geoList.append(Part.Circle(centerPoint, App.Vector(0, 0, 1), diffVec.Length))
    geoIndices = sketch.addGeometry(geoList, construction)

    sketch.setConstruction(geoIndices[-1], True)

    conList = []
    for i in range(0, sides - 1):
        conList.append(Sketcher.Constraint("Coincident", geoIndices[i], 2, geoIndices[i + 1], 1))
    conList.append(Sketcher.Constraint("Coincident", geoIndices[sides - 1], 2, geoIndices[0], 1))
    for i in range(0, sides - 1):
        conList.append(Sketcher.Constraint("Equal", geoIndices[0], geoIndices[i + 1]))
    for i in range(0, sides):
        conList.append(Sketcher.Constraint("PointOnObject", geoIndices[i], 2, geoIndices[-1]))
    sketch.addConstraint(conList)
    return



# ***************************************************************************
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint current density ViewProvider for the document object"
__author__ = "Uwe Stöhr"
__url__ = "https://www.freecad.org"

## @package view_constraint_currentdensity
#  \ingroup FEM
#  \brief view provider for the constraint current density object

from pivy import coin

from femtaskpanels import task_constraint_currentdensity
from . import view_base_femconstraint


class VPConstraintCurrentDensity(view_base_femconstraint.VPBaseFemConstraint):

    def __init__(self, vobj):
        super().__init__(vobj)
        mat = vobj.ShapeAppearance[0]
        mat.DiffuseColor = (0.71, 0.40, 0.11, 0.0)
        vobj.ShapeAppearance = mat

    def setEdit(self, vobj, mode=0):
        return view_base_femconstraint.VPBaseFemConstraint.setEdit(
            self, vobj, mode, task_constraint_currentdensity._TaskPanel
        )

    def attach(self, vobj):
        super().attach(vobj)
        vobj.loadSymbol(self.resource_symbol_dir + "ConstraintCurrentDensity.iv")

    def updateData(self, obj, prop):
        if prop == "Mode":
            symb = obj.ViewObject.SymbolNode.getChild(0)
            if obj.Mode == "Normal":
                obj.ViewObject.RotateSymbol = True
                symb.whichChild.setValue(0)
            elif obj.Mode == "Custom":
                obj.ViewObject.RotateSymbol = False
                symb.whichChild.setValue(-1)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create Array objects.

This includes orthogonal arrays, polar arrays, and circular arrays.
"""
## @package make_array
# \ingroup draftmake
# \brief Provides functions to create Array objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftutils.utils as utils
import draftutils.gui_utils as gui_utils

from draftutils.messages import _wrn, _err
from draftutils.translate import translate
from draftobjects.array import Array

if App.GuiUp:
    from draftviewproviders.view_array import ViewProviderDraftArray
    from draftviewproviders.view_draftlink import ViewProviderDraftLink


def make_array(base_object, arg1, arg2, arg3, arg4=None, arg5=None, arg6=None, use_link=True):
    """Create a Draft Array of the given object.

    Rectangular array
    -----------------
    make_array(object, xvector, yvector, xnum, ynum)
    make_array(object, xvector, yvector, zvector, xnum, ynum, znum)

    xnum of iterations in the x direction
    at xvector distance between iterations, same for y direction
    with yvector and ynum, same for z direction with zvector and znum.

    Polar array
    -----------
    make_array(object, center, totalangle, totalnum) for polar array, or

    center is a vector, totalangle is the angle to cover (in degrees)
    and totalnum is the number of objects, including the original.

    Circular array
    --------------
    make_array(object, rdistance, tdistance, axis, center, ncircles, symmetry)

    In case of a circular array, rdistance is the distance of the
    circles, tdistance is the distance within circles, axis the rotation-axis,
    center the center of rotation, ncircles the number of circles
    and symmetry the number of symmetry-axis of the distribution.

    To Do
    -----
    The `Array` class currently handles three types of arrays,
    orthogonal, polar, and circular. In the future, probably they should be
    split in separate classes so that they are easier to manage.
    """
    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return None

    if use_link:
        # The Array class must be called in this special way
        # to make it a LinkArray
        new_obj = doc.addObject("Part::FeaturePython", "Array", Array(None), None, True)
    else:
        new_obj = doc.addObject("Part::FeaturePython", "Array")
        Array(new_obj)

    new_obj.Base = base_object
    if arg6:
        if isinstance(arg1, (int, float, App.Units.Quantity)):
            new_obj.ArrayType = "circular"
            new_obj.RadialDistance = arg1
            new_obj.TangentialDistance = arg2
            new_obj.Axis = arg3
            new_obj.Center = arg4
            new_obj.NumberCircles = arg5
            new_obj.Symmetry = arg6
        else:
            new_obj.ArrayType = "ortho"
            new_obj.IntervalX = arg1
            new_obj.IntervalY = arg2
            new_obj.IntervalZ = arg3
            new_obj.NumberX = arg4
            new_obj.NumberY = arg5
            new_obj.NumberZ = arg6
    elif arg4:
        new_obj.ArrayType = "ortho"
        new_obj.IntervalX = arg1
        new_obj.IntervalY = arg2
        new_obj.NumberX = arg3
        new_obj.NumberY = arg4
    else:
        new_obj.ArrayType = "polar"
        new_obj.Center = arg1
        new_obj.Angle = arg2
        new_obj.NumberPolar = arg3

    if App.GuiUp:
        if use_link:
            ViewProviderDraftLink(new_obj.ViewObject)
        else:
            if new_obj.ArrayType == "circular":
                new_obj.Proxy.execute(
                    new_obj
                )  # Updates Count which is required for correct DiffuseColor.
            ViewProviderDraftArray(new_obj.ViewObject)
            gui_utils.format_object(new_obj, new_obj.Base)
            new_obj.ViewObject.Proxy.resetColors(new_obj.ViewObject)
        new_obj.Base.ViewObject.hide()
        gui_utils.select(new_obj)

    return new_obj


def makeArray(
    baseobject, arg1, arg2, arg3, arg4=None, arg5=None, arg6=None, name="Array", use_link=False
):
    """Create an Array. DEPRECATED. Use 'make_array'."""
    _wrn(
        "Do not use this function directly; instead, use "
        "'make_ortho_array', 'make_polar_array', "
        "or 'make_circular_array'."
    )

    return make_array(baseobject, arg1, arg2, arg3, arg4, arg5, arg6, use_link)


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2021 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""
Provide tools to access the FreeCAD documentation.

The main usage is using the "show" function. It can retrieve an URL,
a local file (markdown or html), or find a page automatically from
the settings set under Preferences->General->Help.

It doesn't matter what you give, the system will recognize if the contents are
HTML or Markdown and render it appropriately.

Basic usage:

    import Help
    Help.show("Draft Line")
    Help.show("Draft_Line") # works with spaces or underscores
    Help.show("https://wiki.freecad.org/Draft_Line")
    Help.show("https://gitlab.com/freecad/FreeCAD-documentation/-/raw/main/wiki/Draft_Line.md")
    Help.show("/home/myUser/.FreeCAD/Documentation/Draft_Line.md")
    Help.show("http://myserver.com/myfolder/Draft_Line.html")

Preferences keys (in "User parameter:BaseApp/Preferences/Mod/Help"):

    optionBrowser/optionTab/optionDialog (bool): Where to open the help dialog
    optionOnline/optionOffline (bool): where to fetch the documentation from
    URL (string): online location
    Location (string): offline location
    Suffix (string): a suffix to add to the URL, ex: /fr
    StyleSheet (string): optional CSS stylesheet to style the output

Defaults are to open the wiki in the desktop browser
"""

import os
import re
import urllib.request
import urllib.error
import FreeCAD


translate = FreeCAD.Qt.translate
QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP

# texts and icons
WIKI_URL = "https://wiki.freecad.org"
MD_RAW_URL = "https://raw.githubusercontent.com/FreeCAD/FreeCAD-documentation/main/wiki"
MD_RENDERED_URL = "https://github.com/FreeCAD/FreeCAD-documentation/blob/main/wiki"
MD_TRANSLATIONS_FOLDER = "translations"
ERRORTXT = translate(
    "Help",
    "Contents for this page could not be retrieved. Please check settings under menu Edit → Preferences → General → Help",
)
LOCTXT = translate(
    "Help",
    "Help files location could not be determined. Please check settings under menu Edit → Preferences → General → Help",
)
LOGTXT = translate(
    "Help",
    "PySide QtWebEngineWidgets module is not available. Help rendering is done with the system browser",
)
CONVERTTXT = translate(
    "Help",
    "There is no Markdown renderer installed on your system, so this help page is rendered as is. Please install the Markdown or Pandoc Python modules to improve the rendering of this page.",
)
PREFS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Help")
ICON = ":/icons/help-browser.svg"


def show(page, view=None, conv=None):
    """
    show(page,view=None, conv=None):
    Opens a help viewer and shows the given help page.
    The given help page can be a URL pointing to a markdown or HTML file,
    a name page / command name, or a file path pointing to a markdown
    or HTML file. If view is given (an instance of openBrowserHTML.HelpPage or
    any other object with a 'setHtml()' method), the page will be
    rendered there, otherwise a new tab/widget will be created according to
    preferences settings. If conv is given (markdown, pandoc, github, builtin or
    none), the corresponding markdown conversion method is used. Otherwise, the
    module will use the best available.
    In non-GUI mode, this function simply outputs the markdown or HTML text.
    """

    page = underscore_page(page)
    location, _pagename = get_location(page)
    FreeCAD.Console.PrintLog("Help: opening " + location + "\n")
    if not location:
        FreeCAD.Console.PrintError(LOCTXT + "\n")
        return
    md = get_contents(location)
    html = convert(md, conv)
    baseurl = get_uri(location)
    if _pagename != "":
        pagename = _pagename
    else:
        pagename = os.path.basename(page.replace("_", " ").replace(".md", ""))
    title = translate("Help", "Help") + ": " + pagename
    if FreeCAD.GuiUp:
        if PREFS.GetBool("optionTab", False) and get_qtwebwidgets():
            # MDI tab
            show_tab(html, baseurl, title, view)
        elif PREFS.GetBool("optionDialog", False) and get_qtwebwidgets():
            # floating dock window
            show_dialog(html, baseurl, title, view)
        else:
            # desktop web browser - default
            show_browser(location)
    else:
        # console mode, we just print the output
        print(md)


def underscore_page(page):
    """change spaces by underscores in the given page name"""

    if "/" in page:
        page = page.split("/")
        page[-1] = page[-1].replace(" ", "_")
        page = "/".join(page)
    else:
        page.replace(" ", "_")
    return page


def get_uri(location):
    """returns a valid URI from a disk or network location"""

    baseurl = os.path.dirname(location) + "/"
    if baseurl.startswith("/"):  # unix path
        baseurl = "file://" + baseurl
    if baseurl[0].isupper() and (baseurl[1] == ":"):  # windows path
        baseurl = baseurl.replace("\\", "/")
        baseurl = "file:///" + baseurl
    return baseurl


def location_url(url_localized: str, url_english: str) -> tuple:
    """
    Returns localized documentation url and page name, if they exist,
    otherwise defaults to english version.
    Page name is gotten from:
      a) Name/* metadata tag on raw markdown files from github,
         Wiki translators should make sure to add it.
      b) <title> HTML tag
    """
    try:
        req = urllib.request.Request(url_localized)
        with urllib.request.urlopen(req) as response:
            html = response.read().decode("utf-8")
            if url_localized.startswith(MD_RAW_URL):
                pagename_match = re.search(r"Name/.*?:\s*(.+)", html)
            else:
                # Pages from FreeCAD Wiki fall here
                pagename_match = re.search(r"<title>(.*?) - .*?</title>", html)
            if pagename_match is not None:
                return (url_localized, pagename_match.group(1))
            else:
                return (url_localized, "")
    except urllib.error.HTTPError as e:
        return (url_english, "")


def get_location(page) -> tuple:
    """retrieves the location (online or offline) of a given page. Returns the location and the page name"""

    location = ""
    if page.startswith("http"):
        # NOTE: This gets activated when you open a link on the built-in browser, since
        # we don't know the URL, using location_url() fallback to the HTML <title> tag
        return location_url(page, page)
    if page.startswith("file://"):
        return (page[7:], "")
    # offline location
    if os.path.exists(page):
        return (page, "")
    page = page.replace(".md", "")
    page = page.replace(" ", "_")
    page = page.replace("wiki/", "")
    page = page.split("#")[0]
    suffix = PREFS.GetString("Suffix", "")
    pagename = ""
    if suffix:
        if not suffix.startswith("/"):
            suffix = "/" + suffix
    if PREFS.GetBool("optionWiki", True):  # default
        location, pagename = location_url(WIKI_URL + "/" + page + suffix, WIKI_URL + "/" + page)
    elif PREFS.GetBool("optionMarkdown", False):
        if PREFS.GetBool("optionBrowser", False):
            location = MD_RENDERED_URL
        else:
            location = MD_RAW_URL
        if suffix:
            location, pagename = location_url(
                location + "/" + MD_TRANSLATIONS_FOLDER + suffix + "/" + page + ".md",
                location + "/" + page + ".md",
            )
        else:
            location += "/" + page + ".md"
    elif PREFS.GetBool("optionGithub", False):
        location = MD_RENDERED_URL
        if suffix:
            location, pagename = location_url(
                location + "/" + MD_TRANSLATIONS_FOLDER + suffix + "/" + page + ".md",
                location + "/" + page + ".md",
            )
        else:
            location += "/" + page + ".md"
    elif PREFS.GetBool("optionCustom", False):
        location = PREFS.GetString("Location", "")
        if not location:
            location = os.path.join(
                FreeCAD.getUserAppDataDir(),
                "Mod",
                "offline-documentation",
                "FreeCAD-documentation-main",
                "wiki",
            )
        for ext in (".md", ".html", ".htm"):
            if os.path.exists(os.path.join(location, page + ext)):
                location = os.path.join(location, page + ext)
                break
    return (location, pagename)


def show_browser(url):
    """opens the desktop browser with the given URL"""

    from PySide import QtCore, QtGui

    try:
        ret = QtGui.QDesktopServices.openUrl(QtCore.QUrl(url))
        if not ret:
            # some users reported problems with the above
            import webbrowser

            webbrowser.open_new(url)
    except:
        import webbrowser

        webbrowser.open_new(url)


def show_dialog(html, baseurl, title, view=None):
    """opens a dock dialog with the given html"""

    from PySide import QtCore

    if view:  # reusing existing view
        view.setHtml(html, baseUrl=QtCore.QUrl(baseurl))
        view.parent().parent().setWindowTitle(title)
    else:
        openBrowserHTML(html, baseurl, title, ICON, dialog=True)


def show_tab(html, baseurl, title, view=None):
    """opens a MDI tab with the given html"""

    from PySide import QtCore

    if view:  # reusing existing view
        view.setHtml(html, baseUrl=QtCore.QUrl(baseurl))
        view.parent().parent().setWindowTitle(title)
    else:
        openBrowserHTML(html, baseurl, title, ICON)


def get_qtwebwidgets():
    """verifies if qtwebengine is available"""

    try:
        from PySide import QtWebEngineWidgets
    except:
        FreeCAD.Console.PrintLog(LOGTXT + "\n")
        return False
    else:
        return True


def get_contents(location):
    """retrieves text contents of a given page"""

    import urllib

    if location.startswith("http"):
        import urllib.request

        try:
            r = urllib.request.urlopen(location)
        except:
            return ERRORTXT
        contents = r.read().decode("utf8")
        return contents
    else:
        if os.path.exists(location):
            with open(location, mode="r", encoding="utf8") as f:
                contents = f.read()
            return contents
    return ERRORTXT


def convert(content, force=None):
    """converts the given markdown code to html. Force can be None (automatic)
    or markdown, pandoc, github or raw/builtin"""

    import urllib

    def convert_markdown(m):
        try:
            import markdown
            from markdown.extensions import codehilite

            return markdown.markdown(m, extensions=["codehilite"])
        except:
            return None

    def convert_pandoc(m):
        try:
            import pypandoc

            return pypandoc.convert_text(m, "html", format="md")
        except:
            return None

    def convert_github(m):
        try:
            import json
            import urllib.request

            data = {"text": m, "mode": "markdown"}
            bdata = json.dumps(data).encode("utf-8")
            return (
                urllib.request.urlopen("https://api.github.com/markdown", data=bdata)
                .read()
                .decode("utf8")
            )
        except:
            return None

    def convert_raw(m):
        # simple and dirty regex-based markdown to html

        import re

        f = re.DOTALL | re.MULTILINE
        m = re.sub(r"^##### (.*?)\n", r"<h5>\1</h5>\n", m, flags=f)  # ##### titles
        m = re.sub(r"^#### (.*?)\n", r"<h4>\1</h4>\n", m, flags=f)  # #### titles
        m = re.sub(r"^### (.*?)\n", r"<h3>\1</h3>\n", m, flags=f)  # ### titles
        m = re.sub(r"^## (.*?)\n", r"<h2>\1</h2>\n", m, flags=f)  # ## titles
        m = re.sub(r"^# (.*?)\n", r"<h1>\1</h1>\n", m, flags=f)  # # titles
        m = re.sub(r"!\[(.*?)\]\((.*?)\)", r'<img alt="\1" src="\2">', m, flags=f)  # images
        m = re.sub(r"\[(.*?)\]\((.*?)\)", r'<a href="\2">\1</a>', m, flags=f)  # links
        m = re.sub(r"\*\*(.*?)\*\*", r"<b>\1</b>", m)  # bold
        m = re.sub(r"\*(.*?)\*", r"<i>\1</i>", m)  # italic
        m = re.sub(r"\n\n", r"<br/>", m, flags=f)  # double new lines
        m += "\n<br/><hr/><small>" + CONVERTTXT + "</small>"
        return m

    if "<html" in content:
        # this is html already
        return content

    if force == "markdown":
        html = convert_markdown(content)
    elif force == "pandoc":
        html = convert_pandoc(content)
    elif force == "github":
        html = convert_github(content)
    elif force in ["raw", "builtin"]:
        html = convert_raw(content)
    elif force == "none":
        return content
    else:
        # auto mode
        html = convert_pandoc(content)
        if not html:
            html = convert_markdown(content)
            if not html:
                html = convert_raw(content)
    if not "<html" in html:
        html = (
            '<html>\n<head>\n<meta charset="utf-8"/>\n</head>\n<body>\n\n'
            + html
            + "</body>\n</html>"
        )
    # insert css
    css = None
    cssfile = PREFS.GetString("StyleSheet", "")
    if not cssfile:
        cssfile = os.path.join(os.path.dirname(__file__), "default.css")
    if False:  # linked CSS file
        # below code doesn't work in FreeCAD apparently because it prohibits cross-URL stuff
        cssfile = urllib.parse.urljoin("file:", urllib.request.pathname2url(cssfile))
        css = '<link rel="stylesheet" type="text/css" href="' + cssfile + '"/>'
    else:
        if os.path.exists(cssfile):
            with open(cssfile) as cf:
                css = cf.read()
            if css:
                css = "<style>\n" + css + "\n</style>"
        else:
            print("Debug: Help: Unable to open css file:", cssfile)
    if css:
        html = html.replace("</head>", css + "\n</head>")
    return html


def add_preferences_page():
    """adds the Help preferences page to the UI"""

    import FreeCADGui

    page = os.path.join(os.path.dirname(__file__), "dlgPreferencesHelp.ui")
    FreeCADGui.addPreferencePage(page, QT_TRANSLATE_NOOP("QObject", "General"))


def add_language_path():
    """registers the Help translations to FreeCAD"""

    import FreeCADGui
    import Help_rc

    FreeCADGui.addLanguagePath(":/translations")


def openBrowserHTML(html, baseurl, title, icon, dialog=False):
    """creates a browser view and adds it as a FreeCAD MDI tab or dockable dialog"""

    import FreeCADGui
    from PySide import QtCore, QtGui, QtWidgets, QtWebEngineWidgets

    # turn an int into a qt dock area
    def getDockArea(area):
        if area == 1:
            return QtCore.Qt.LeftDockWidgetArea
        elif area == 4:
            return QtCore.Qt.TopDockWidgetArea
        elif area == 8:
            return QtCore.Qt.BottomDockWidgetArea
        else:
            return QtCore.Qt.RightDockWidgetArea

    # save dock widget size and location
    def onDockLocationChanged(area):
        PREFS.SetInt("dockWidgetArea", int(area.value))
        mw = FreeCADGui.getMainWindow()
        dock = mw.findChild(QtWidgets.QDockWidget, "HelpWidget")
        if dock:
            PREFS.SetBool("dockWidgetFloat", dock.isFloating())
            PREFS.SetInt("dockWidgetWidth", dock.width())
            PREFS.SetInt("dockWidgetHeight", dock.height())

    # a custom page that handles .md links
    class HelpPage(QtWebEngineWidgets.QWebEnginePage):
        def acceptNavigationRequest(self, url, _type, isMainFrame):
            if _type == QtWebEngineWidgets.QWebEnginePage.NavigationTypeLinkClicked:
                show(url.toString(), view=self)
            return super().acceptNavigationRequest(url, _type, isMainFrame)

    mw = FreeCADGui.getMainWindow()
    view = QtWebEngineWidgets.QWebEngineView()
    page = HelpPage(None, view)
    page.setHtml(html, baseUrl=QtCore.QUrl(baseurl))
    view.setPage(page)

    if dialog:
        area = PREFS.GetInt("dockWidgetArea", 2)
        floating = PREFS.GetBool("dockWidgetFloat", True)
        height = PREFS.GetBool("dockWidgetWidth", 200)
        width = PREFS.GetBool("dockWidgetHeight", 300)
        dock = mw.findChild(QtWidgets.QDockWidget, "HelpWidget")
        if not dock:
            dock = QtWidgets.QDockWidget()
            dock.setObjectName("HelpWidget")
            mw.addDockWidget(getDockArea(area), dock)
            dock.setFloating(floating)
            dock.setGeometry(dock.x(), dock.y(), width, height)
            dock.dockLocationChanged.connect(onDockLocationChanged)
        dock.setWidget(view)
        dock.setWindowTitle(title)
        dock.setWindowIcon(QtGui.QIcon(icon))
        dock.show()
    else:
        mdi = mw.findChild(QtWidgets.QMdiArea)
        sw = mdi.addSubWindow(view)
        sw.setWindowTitle(title)
        sw.setWindowIcon(QtGui.QIcon(icon))
        sw.show()
        mdi.setActiveSubWindow(sw)



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileNotice: Part of the AddonManager.


import addonmanager_freecad_interface as fci

try:
    from PySide.QtCore import QT_TRANSLATE_NOOP
except ImportError:
    try:
        from PySide6.QtCore import QT_TRANSLATE_NOOP
    except ImportError:
        from PySide2.QtCore import QT_TRANSLATE_NOOP


class ToolbarAdapter:
    """When run within FreeCAD, interface with the toolbar preferences and command-creation
    code to actually allow the modification of toolbars (e.g., to install a button that runs a
    Macro when clicked). Outside FreeCAD, calling any of these methods raises an exception.
    Test code is expected to Mock this class."""

    params = None

    def __init__(self):
        if fci.FreeCAD is None:
            raise RuntimeError("ToolbarAdapter can only be used when run from within FreeCAD")
        if self.params is None:
            self.params = fci.FreeCAD.ParamGet("User parameter:BaseApp/Workbench/Global/Toolbar")

    def get_toolbars(self):
        """Get a list of toolbars: the result is a set of parameter groups, each representing a toolbar."""
        toolbars = []
        toolbar_groups = self.params.GetGroups()
        for group in toolbar_groups:
            toolbar = self.params.GetGroup(group)
            toolbars.append(toolbar)
        return toolbars

    def create_new_custom_toolbar(self):
        """Create a new custom toolbar and returns its preference group."""

        # We need two names: the name of the auto-created toolbar, as it will be displayed to the
        # user in various menus, and the underlying name of the toolbar group. Both must be
        # unique.

        # First, the displayed name
        custom_toolbar_name = str(QT_TRANSLATE_NOOP("Workbench", "Auto-Created Macro Toolbar"))
        custom_toolbars = self.params.GetGroups()
        name_taken = self.check_for_toolbar(custom_toolbar_name)
        if name_taken:
            i = 2  # Don't use (1), start at (2)
            while True:
                test_name = custom_toolbar_name + f" ({i})"
                if not self.check_for_toolbar(test_name):
                    custom_toolbar_name = test_name
                    break
                i = i + 1

        # Second, the toolbar preference group name
        i = 1
        while True:
            new_group_name = "Custom_" + str(i)
            if new_group_name not in custom_toolbars:
                break
            i = i + 1

        custom_toolbar = self.params.GetGroup(new_group_name)
        custom_toolbar.SetString("Name", custom_toolbar_name)
        custom_toolbar.SetBool("Active", True)
        return custom_toolbar

    def check_for_toolbar(self, toolbar_name: str) -> bool:
        """Returns True if the toolbar exists, otherwise False"""
        return self.get_toolbar_with_name(toolbar_name) is not None

    def get_toolbar_with_name(self, name: str):
        """Try to find a toolbar with a given name. Returns the preference group for the toolbar
        if found, or None if it does not exist."""
        custom_toolbars = self.params.GetGroups()
        for toolbar in custom_toolbars:
            group = self.params.GetGroup(toolbar)
            group_name = group.GetString("Name", "")
            if group_name == name:
                return group
        return None

    @staticmethod
    def create_custom_command(
        toolbar,
        filename,
        menu_text,
        tooltip_text,
        whats_this_text,
        status_tip_text,
        pixmap_text,
    ):
        """Create a custom command and reload the active workbench so that toolbars get recreated."""
        command_name = fci.FreeCADGui.Command.createCustomCommand(
            filename, menu_text, tooltip_text, whats_this_text, status_tip_text, pixmap_text
        )
        toolbar.SetString(command_name, "FreeCAD")

        # Force the toolbars to be recreated
        wb = fci.FreeCADGui.activeWorkbench()
        wb.reloadActive()

    def remove_custom_toolbar_button(self, filename):
        """Given a macro filename, remove an associated toolbar button."""
        command = fci.FreeCADGui.Command.findCustomCommand(filename)
        if not command:
            return
        toolbars = self.get_toolbars()
        for toolbar in toolbars:
            if toolbar.GetString(command, "*") != "*":
                toolbar.RemString(command)

        fci.FreeCADGui.Command.removeCustomCommand(command)

        # Force the toolbars to be recreated
        wb = fci.FreeCADGui.activeWorkbench()
        wb.reloadActive()

    @staticmethod
    def get_toolbar_name(group):
        """Given a toolbar preference group, return the name of the toolbar."""
        return group.GetString("Name", "")

    def get_custom_toolbars(self):
        """Get a list of toolbar preference groups"""
        return self.params.GetGroups()

    @staticmethod
    def find_custom_command(filename: str):
        """Wrap calls to `FreeCADGui.Command.findCustomCommand` so it can be faked in testing."""
        return fci.FreeCADGui.Command.findCustomCommand(filename)
